# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["DataStore"]

import typing as t

T = t.TypeVar("T")


class DataStore(t.Dict[str, t.Any]):
    """
    Data storage class allowing setting, retrieval and unsetting of custom
    attributes. This class subclasses :obj:`dict` so the data can be accessed the same
    as you would a dictionary as well as using dot notation.

    Example:

        .. code-block:: python

            >>> d = DataStore()
            >>> d.foo = "bar"
            >>> d.foo
            'bar'
            >>> d["foo"]
            'bar'
            >>> d
            DataStore(foo='bar')
            >>> d.pop("foo")
            'bar'
            >>> d
            DataStore()
    """

    def __repr__(self) -> str:
        return "DataStore(" + ", ".join(f"{k}={v!r}" for k, v in self.items()) + ")"

    def __getattr__(self, item: str) -> t.Any:
        return self.get(item)

    def __setattr__(self, key: str, value: t.Any) -> None:
        self[key] = value

    def __delattr__(self, item: str) -> None:
        self.pop(item, None)

    def get_as(self, item: str, type: t.Type[T]) -> T:
        """
        Helper method to allow type-complete getting of items from this ``DataStore``.

        Args:
            item (:obj:`str`): The name of the key that the item is stored at.
            type (Type[T]): The type to cast the item as.

        Returns:
            T: The item stored at key ``item``, cast to the given type.

        Raises:
            :obj:`ValueError`: If a key of name ``item`` has not been set.

        Note:
            This does **not** verify types, it just performs a :meth:`typing.cast` to fool the
            type system into thinking that the return value is of the correct type.

        .. versionadded:: 2.2.4
        """
        if item not in self:
            raise TypeError(f"Item {item!r} was never set")
        return t.cast(T, self.get(item))



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["StringPaginator", "EmbedPaginator", "Paginator"]

import abc
import io
import textwrap
import typing as t

import hikari

T = t.TypeVar("T")


class Paginator(abc.ABC, t.Generic[T]):
    @abc.abstractmethod
    def __init__(
        self,
        *,
        max_lines: t.Optional[int] = None,
        max_chars: int = 2000,
        prefix: str = "",
        suffix: str = "",
        line_separator: str = "\n",
        page_factory: t.Callable[[int, str], T] = lambda i, s: s,  # type: ignore
    ) -> None:
        self._page_prefix: str = prefix
        self._page_suffix: str = suffix
        # Dummy minimum content size. This is never used, but is just symbolic for readability.
        min_content = f"A{line_separator}"

        # at least one line of content.
        extra_lines = prefix.count(line_separator) + suffix.count(line_separator)
        min_total_lines = min_content.count(line_separator) + extra_lines

        if max_lines is not None:
            if max_lines < min_total_lines:
                raise ValueError(f"This configuration requires at least {min_total_lines} lines per page!")

        # At least 1 character per page, or we recurse forever!
        prefix_len = len(prefix) + len(suffix)
        min_total_len = prefix_len + len(min_content)

        if max_chars < min_total_len:
            raise ValueError(f"This configuration requires at least {min_total_len} characters per page.")

        self._max_total_chars = max_chars
        self._max_total_lines = max_lines if max_lines is not None else float("inf")
        self._max_content_chars = max_chars - prefix_len
        self._max_content_lines = max_lines - extra_lines if max_lines is not None else float("inf")
        self._line_separator = line_separator
        self._next_page: io.StringIO = io.StringIO()
        self._pages: t.List[str] = []
        self._page_factory = page_factory
        self.current_page: int = 0

    def __len__(self) -> int:
        return len(self._pages)

    def build_pages(self, page_number_start: int = 1) -> t.Iterator[T]:
        """
        The current pages that have been created.

        Args:
            page_number_start (:obj:`int`): The page number to start at.
                Defaults to ``1``.

        Returns:
            Iterator[ T ]: Lazy generator of each page.
        """
        # Only add the last page if it is not empty.
        if self._next_page.tell():
            last_page = self._next_page.getvalue()
            if len(last_page) > 0:
                self.new_page()

        for i, page in enumerate(self._pages, start=page_number_start):
            yield self._page_factory(i, page)

    def add_line(self, line: t.Any) -> None:
        """
        Add a line to the paginator.

        Args:
            line (Any): The line to add to the paginator.
                Will be converted to a :obj:`str`.

        Returns:
            ``None``
        """
        whole_text = str(line).replace("\t", (" " * 4)).replace("\r", "").split(self._line_separator)

        for line in whole_text:
            self._add_one_line(line)

    def _add_one_line(self, line: str) -> None:
        existing_chars, existing_lines = self._sizes()
        remaining_chars = self._max_content_chars - existing_chars
        remaining_lines = self._max_content_lines - existing_lines
        this_char_count = len(line)

        if not self._next_page.tell():
            self._next_page.write(self._page_prefix)

        if this_char_count > self._max_content_chars:
            self._chunk_add(line)
            return

        if remaining_chars < this_char_count or remaining_lines <= 0:
            self.new_page()
            self._add_one_line(line)
            return

        self._next_page.write(line)
        self._next_page.write(self._line_separator)

    def _chunk_add(self, line: str) -> None:
        # Try to split up words, if not, break mid-word.
        wrapper = textwrap.TextWrapper(
            width=self._max_content_chars,
            expand_tabs=True,
            tabsize=4,
            max_lines=self._max_content_lines,  # type: ignore
        )

        lines = wrapper.wrap(line)

        for line in lines:
            if len(line) > self._max_content_chars - len(self._line_separator):
                for i in range(0, len(line), self._max_content_chars):
                    next_line = line[i : i + self._max_content_chars]
                    self.add_line(next_line)
            else:
                self.add_line(line)

    def new_page(self) -> None:
        """
        Start a new page.

        Returns:
            ``None``
        """
        # Remove final newline if it is there.
        next_page = self._next_page.getvalue()
        if next_page.endswith(self._line_separator):
            next_page = next_page[: -len(self._line_separator)]
        next_page += self._page_suffix

        # Append page
        self._pages.append(next_page)

        # Clear buffer
        self._next_page.seek(0, 0)
        self._next_page.truncate(0)

    def _sizes(self) -> t.Tuple[int, int]:
        page = self._next_page.getvalue()[len(self._page_prefix) :]
        current_chars = len(page)
        current_lines = page.count(self._line_separator)
        return current_chars, current_lines


class StringPaginator(Paginator[str]):
    """
    Creates pages from lines of text according to the given parameters.

    Text should be added to the paginator using :meth:`~.utils.pag.StringPaginator.add_line`, which
    will then be split up into an appropriate number of pages, accessible
    through :attr:`~.utils.pag.StringPaginator.pages`.

    Keyword Args:
        max_lines (Optional[ :obj:`int` ]): The maximum number of lines per page. Defaults to ``None``, meaning
            pages will use the ``max_chars`` param instead.
        max_chars (:obj:`int`): The maximum number of characters per page. Defaults to ``2000``, the max character
            limit for a discord message.
        prefix (:obj:`str`): The string to prefix every page with. Defaults to an empty string.
        suffix (:obj:`str`): The string to suffix every page with. Defaults to an empty string.

    Example:
        An example command using pagination to display all the guilds the bot is in.

        .. code-block:: python

            from lightbulb.utils.pag import StringPaginator

            @bot.command()
            async def guilds(ctx):
                guilds = await bot.rest.fetch_my_guilds()
                pag = StringPaginator(max_lines=10)
                for n, guild in enumerate(guilds, start=1):
                    pag.add_line(f"**{n}.** {guild.name}")
                for page in pag.build_pages():
                    await ctx.respond(page)
    """

    def __init__(
        self,
        *,
        max_lines: t.Optional[int] = None,
        max_chars: int = 2000,
        prefix: str = "",
        suffix: str = "",
        line_separator: str = "\n",
    ) -> None:
        super().__init__(
            max_lines=max_lines,
            max_chars=max_chars,
            prefix=prefix,
            suffix=suffix,
            line_separator=line_separator,
        )


class EmbedPaginator(Paginator[hikari.Embed]):
    """
    Creates embed pages from lines of text according to the given parameters.

    Text is added to the paginator the same way as :obj:`~.utils.pag.StringPaginator`. The paginated text will
    be run though the defined :meth:`embed_factory`, or if no embed factory is defined
    then it will be inserted into the description of a default embed.

    Keyword Args:
        max_lines (Optional[ :obj:`int` ]): The maximum number of lines per page. Defaults to ``None``, meaning
            pages will use the ``max_chars`` param instead.
        max_chars (:obj:`int`): The maximum number of characters per page. Defaults to ``2048``, the max character
            limit for a discord message.
        prefix (:obj:`str`): The string to prefix every page with. Defaults to an empty string.
        suffix (:obj:`str`): The string to suffix every page with. Defaults to an empty string.
    """

    def __init__(
        self,
        *,
        max_lines: t.Optional[int] = None,
        max_chars: int = 2048,
        prefix: str = "",
        suffix: str = "",
        line_separator: str = "\n",
    ) -> None:
        super().__init__(
            max_lines=max_lines,
            max_chars=max_chars,
            prefix=prefix,
            suffix=suffix,
            line_separator=line_separator,
            page_factory=lambda i, s: hikari.Embed(description=s).set_footer(text=f"Page {i}"),
        )

    def embed_factory(self) -> t.Callable[[t.Callable[[int, str], hikari.Embed]], t.Callable[[int, str], hikari.Embed]]:
        """
        A decorator to mark a function as the paginator's embed factory. The page index and page content
        will be passed to the function when a new page is to be created.

        Example:
            The following code will give each embed created a random colour.

            .. code-block:: python

                from random import randint
                from lightbulb.utils.pag import EmbedPaginator
                from hikari import Embed

                pag = EmbedPaginator()

                @pag.embed_factory()
                def build_embed(page_index, page_content):
                    return Embed(description=page_content, colour=randint(0, 0xFFFFFF))

        See Also:
            :meth:`set_embed_factory`
        """

        def decorate(func: t.Callable[[int, str], hikari.Embed]) -> t.Callable[[int, str], hikari.Embed]:
            self.set_embed_factory(func)
            return func

        return decorate

    def set_embed_factory(self, func: t.Callable[[int, str], hikari.Embed]) -> None:
        """
        Method to set a callable as the paginator's embed factory. Alternative to :meth:`embed_factory`.

        Args:
            func (Callable[ [ :obj:`int`, :obj:`str` ], :obj:`~hikari.embeds.Embed` ]): The callable to
                set as the paginator's embed factory.

        Returns:
            ``None``

        See Also:
            :meth:`embed_factory`
        """
        self._page_factory = func



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["MessageContext"]

import typing as t

import hikari

from lightbulb import commands
from lightbulb.context import base

if t.TYPE_CHECKING:
    from lightbulb import app as app_


class MessageContext(base.ApplicationContext):
    """
    An implementation of :obj:`~.context.base.Context` for message context menu commands.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the context is linked to.
        event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): The event to create the context
            from.
        command (:obj:`~.commands.message.MessageCommand`): The command that the context is for.

    Note that the target message that this command was invoked on will **always** be stored
    as the option ``target``.

    Example:

        .. code-block:: python

            @lightbulb.command("id", "Gets the ID of the selected message")
            @lightbulb.implements(lightbulb.MessageCommand)
            async def get_message_id(ctx: lightbulb.MessageContext) -> None:
                await ctx.respond(ctx.options.target.id)

    .. versionadded:: 2.2.0
    """

    __slots__ = ("_options",)

    def __init__(
        self, app: app_.BotApp, event: hikari.InteractionCreateEvent, command: commands.message.MessageCommand
    ) -> None:
        super().__init__(app, event, command)
        assert self.resolved is not None
        assert self.interaction.target_id is not None
        self._options = {"target": self.resolved.messages.get(self.interaction.target_id)}

    @property
    def raw_options(self) -> t.Dict[str, t.Any]:
        return self._options

    @property
    def command(self) -> commands.message.MessageCommand:
        assert isinstance(self._command, commands.message.MessageCommand)
        return self._command

    @property
    def prefix(self) -> str:
        return "\N{THREE BUTTON MOUSE}\N{VARIATION SELECTOR-16}"



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "Check",
    "owner_only",
    "guild_only",
    "dm_only",
    "bot_only",
    "webhook_only",
    "human_only",
    "nsfw_channel_only",
    "has_roles",
    "has_role_permissions",
    "has_channel_permissions",
    "has_guild_permissions",
    "bot_has_guild_permissions",
    "bot_has_role_permissions",
    "bot_has_channel_permissions",
    "has_attachments",
]

import functools
import inspect
import operator
import typing as t
import warnings

import hikari

from lightbulb import context as context_
from lightbulb import errors
from lightbulb.utils import permissions

if t.TYPE_CHECKING:
    from lightbulb import app
    from lightbulb import commands
    from lightbulb import plugins

_CallbackT = t.Union[
    t.Callable[[context_.base.Context], t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]], functools.partial
]


class _ExclusiveCheck:
    def __init__(self, *checks: "Check") -> None:
        self._checks = list(checks)

    def __name__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"ExclusiveCheck({', '.join(repr(c) for c in self._checks)})"

    def __or__(self, other: t.Union["_ExclusiveCheck", "Check"]) -> "_ExclusiveCheck":
        if isinstance(other, _ExclusiveCheck):
            self._checks.extend(other._checks)
        else:
            self._checks.append(other)
        return self

    async def _evaluate(self, context: context_.base.Context) -> bool:
        failed = []

        for check in self._checks:
            try:
                res = check(context)
                if inspect.iscoroutine(res):
                    res = await res

                if not res:
                    raise errors.CheckFailure(f"Check {check.__name__!r} failed")
                return True
            except Exception as ex:
                if isinstance(ex, errors.CheckFailure) and not ex.__cause__:
                    ex = errors.CheckFailure(str(ex), causes=[ex])
                    ex.__cause__ = ex
                failed.append(ex)

        if failed:
            if len(failed) == 1:
                raise failed[0]
            raise errors.CheckFailure(
                "None of the exclusive checks passed: " + ", ".join(str(ex) for ex in failed), causes=failed
            )
        return True

    def __call__(self, context: context_.base.Context) -> t.Coroutine[t.Any, t.Any, bool]:
        return self._evaluate(context)

    def add_to_object_hook(
        self, obj: t.Union[plugins.Plugin, app.BotApp, commands.base.CommandLike]
    ) -> t.Union[plugins.Plugin, app.BotApp, commands.base.CommandLike]:
        for check in self._checks:
            check.add_to_object_hook(obj)
        return obj


class Check:
    """
    Class representing a check. Check functions can be synchronous or asynchronous functions which take
    a single argument, which will be the context that the command is being invoked under, and return
    a boolean or raise a :obj:`.errors.CheckFailure` indicating whether the check passed or failed.

    Args:
        p_callback (CallbackT): Check function to use for prefix commands.
        s_callback (Optional[CallbackT]): Check function to use for slash commands.
        m_callback (Optional[CallbackT]): Check function to use for message commands.
        u_callback (Optional[CallbackT]): Check function to use for user commands.
        add_hook (Optional[Callable[[T], T]]): Function called when the check is added to an object.
    """

    __slots__ = ("prefix_callback", "slash_callback", "message_callback", "user_callback", "add_to_object_hook")

    def __init__(
        self,
        p_callback: _CallbackT,
        s_callback: t.Optional[_CallbackT] = None,
        m_callback: t.Optional[_CallbackT] = None,
        u_callback: t.Optional[_CallbackT] = None,
        add_hook: t.Optional[
            t.Callable[
                [t.Union[plugins.Plugin, app.BotApp, commands.base.CommandLike]],
                t.Union[plugins.Plugin, app.BotApp, commands.base.CommandLike],
            ]
        ] = None,
    ) -> None:
        self.prefix_callback = p_callback
        self.slash_callback = s_callback or p_callback
        self.message_callback = m_callback or p_callback
        self.user_callback = u_callback or p_callback
        self.add_to_object_hook = add_hook or (lambda o: o)

    def __repr__(self) -> str:
        return f"Check({self.__name__.strip('_')})"

    def __or__(self, other: t.Union["Check", _ExclusiveCheck]) -> _ExclusiveCheck:
        return _ExclusiveCheck(self) | other

    @property
    def __name__(self) -> str:
        if isinstance(self.prefix_callback, functools.partial):
            return self.prefix_callback.func.__name__
        return self.prefix_callback.__name__

    def __call__(self, context: context_.base.Context) -> t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]:
        if isinstance(context, context_.prefix.PrefixContext):
            return self.prefix_callback(context)
        elif isinstance(context, context_.slash.SlashContext):
            return self.slash_callback(context)
        elif isinstance(context, context_.message.MessageContext):
            return self.message_callback(context)
        elif isinstance(context, context_.user.UserContext):
            return self.user_callback(context)
        return True


async def _owner_only(context: context_.base.Context) -> bool:
    if not context.app.owner_ids:
        context.app.owner_ids = await context.app.fetch_owner_ids()

    if context.author.id not in context.app.owner_ids:
        raise errors.NotOwner("You are not the owner of this bot")
    return True


def _guild_only(context: context_.base.Context) -> bool:
    if context.guild_id is None:
        raise errors.OnlyInGuild("This command can only be used in a guild")
    return True


def _dm_only(context: context_.base.Context) -> bool:
    if context.guild_id is not None:
        raise errors.OnlyInDM("This command can only be used in DMs")
    return True


def _bot_only(context: context_.base.Context) -> bool:
    if not context.author.is_bot:
        raise errors.BotOnly("This command can only be used by bots")
    return True


def _webhook_only(context: context_.base.Context) -> bool:
    if not isinstance(context, context_.prefix.PrefixContext):
        raise errors.WebhookOnly("This command can only be used by webhooks")
    if context.event.message.webhook_id is None:
        raise errors.WebhookOnly("This command can only be used by webhooks")
    return True


def _human_only(context: context_.base.Context) -> bool:
    if isinstance(context, context_.prefix.PrefixContext):
        if context.author.is_bot or context.event.message.webhook_id is not None:
            raise errors.HumanOnly("This command can only be used by humans")
    if context.author.is_bot:
        raise errors.HumanOnly("This command can only be used by humans")
    return True


def _nsfw_channel_only(context: context_.base.Context) -> bool:
    if context.guild_id is None:
        raise errors.NSFWChannelOnly("This command can only be used in NSFW channels")

    channel = context.get_channel()

    # Threads do not have their nsfw status attached to them,
    # they inherit that from the channel they are part of
    if isinstance(channel, hikari.GuildThreadChannel):
        channel = context.app.cache.get_guild_channel(channel.parent_id)

    assert isinstance(channel, hikari.PermissibleGuildChannel)

    if not channel.is_nsfw:
        raise errors.NSFWChannelOnly("This command can only be used in NSFW channels")
    return True


def _has_roles(
    context: context_.base.Context, *, roles: t.Sequence[int], check_func: t.Callable[[t.Sequence[bool]], bool]
) -> bool:
    _guild_only(context)
    assert context.member is not None

    missing_roles = [r for r in roles if r not in context.member.role_ids]
    # TODO: could potentially optimise any() by stopping the iteration when a required role was found
    if (check_func is all and missing_roles) or len(missing_roles) == len(roles):
        raise errors.MissingRequiredRole(
            "You are missing one or more roles required in order to run this command",
            roles=missing_roles,
            mode=check_func,
        )
    return True


def _has_guild_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    if isinstance(context, context_.base.ApplicationContext):
        cmd = t.cast("commands.base.ApplicationCommand", context.invoked)
        if (context.guild_id is None and cmd.app_command_dm_enabled) or cmd.app_command_bypass_author_permission_checks:
            return True

    _guild_only(context)

    if isinstance(context, context_.base.ApplicationContext):
        assert context.interaction.member is not None
        missing_perms = ~context.interaction.member.permissions & perms
        if missing_perms is not hikari.Permissions.NONE:
            raise errors.MissingRequiredPermission(
                "You are missing one or more permissions required in order to run this command", perms=missing_perms
            )
        return True

    channel = context.get_channel()

    # Threads do not have permissions attached to them, they inherit that from
    # the channel they are part of
    if isinstance(channel, hikari.GuildThreadChannel):
        channel = context.app.cache.get_guild_channel(channel.parent_id)

    if channel is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    assert context.member is not None and isinstance(channel, hikari.PermissibleGuildChannel)
    missing_perms = ~permissions.permissions_in(channel, context.member) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.MissingRequiredPermission(
            "You are missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _has_role_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    if isinstance(context, context_.base.ApplicationContext):
        cmd = t.cast("commands.base.ApplicationCommand", context.invoked)
        if (context.guild_id is None and cmd.app_command_dm_enabled) or cmd.app_command_bypass_author_permission_checks:
            return True

    _guild_only(context)

    assert context.member is not None
    missing_perms = ~permissions.permissions_for(context.member) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.MissingRequiredPermission(
            "You are missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _has_channel_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    if isinstance(context, context_.base.ApplicationContext):
        cmd = t.cast("commands.base.ApplicationCommand", context.invoked)
        if (context.guild_id is None and cmd.app_command_dm_enabled) or cmd.app_command_bypass_author_permission_checks:
            return True

    _guild_only(context)

    channel = context.get_channel()

    # Threads do not have permissions attached to them, they inherit that from
    # the channel they are part of
    if isinstance(channel, hikari.GuildThreadChannel):
        channel = context.app.cache.get_guild_channel(channel.parent_id)

    if channel is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    assert context.member is not None and isinstance(channel, hikari.PermissibleGuildChannel)
    missing_perms = ~permissions.permissions_in(channel, context.member, include_guild_permissions=False) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.MissingRequiredPermission(
            "You are missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _bot_has_guild_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    _guild_only(context)

    # TODO - do we need to check for dm_enabled for these checks?
    if isinstance(context, context_.base.ApplicationContext):
        assert context.interaction.app_permissions is not None
        missing_perms = ~context.interaction.app_permissions & perms
        if missing_perms is not hikari.Permissions.NONE:
            raise errors.BotMissingRequiredPermission(
                "The bot is missing one or more permissions required in order to run this command", perms=missing_perms
            )
        return True

    channel = context.get_channel()

    # Threads do not have permissions attached to them, they inherit that from
    # the channel they are part of
    if isinstance(channel, hikari.GuildThreadChannel):
        channel = context.app.cache.get_guild_channel(channel.parent_id)

    if channel is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    me = context.app.get_me()
    assert me is not None
    assert context.guild_id is not None  # Asserted above through _guild_only
    member = context.app.cache.get_member(context.guild_id, me.id)
    if member is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    assert isinstance(channel, hikari.PermissibleGuildChannel)
    missing_perms = ~permissions.permissions_in(channel, member) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.BotMissingRequiredPermission(
            "The bot is missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _bot_has_role_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    _guild_only(context)

    me = context.app.get_me()
    assert me is not None
    assert context.guild_id is not None  # Asserted above through _guild_only
    member = context.app.cache.get_member(context.guild_id, me.id)
    if member is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    missing_perms = ~permissions.permissions_for(member) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.BotMissingRequiredPermission(
            "The bot is missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _bot_has_channel_permissions(context: context_.base.Context, *, perms: hikari.Permissions) -> bool:
    _guild_only(context)

    channel = context.get_channel()

    # Threads do not have permissions attached to them, they inherit that from
    # the channel they are part of
    if isinstance(channel, hikari.GuildThreadChannel):
        channel = context.app.cache.get_guild_channel(channel.parent_id)

    if channel is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    me = context.app.get_me()
    assert me is not None
    assert context.guild_id is not None  # Asserted above through _guild_only
    member = context.app.cache.get_member(context.guild_id, me.id)

    if member is None:
        raise errors.InsufficientCache("Some objects required for this check could not be resolved from the cache")

    assert isinstance(channel, hikari.PermissibleGuildChannel)
    missing_perms = ~permissions.permissions_in(channel, member, include_guild_permissions=False) & perms
    if missing_perms is not hikari.Permissions.NONE:
        raise errors.BotMissingRequiredPermission(
            "The bot is missing one or more permissions required in order to run this command", perms=missing_perms
        )
    return True


def _has_attachments(context: context_.base.Context, *, file_exts: t.Sequence[str] = ()) -> bool:
    if not context.attachments:
        raise errors.MissingRequiredAttachment("Missing attachment(s) required to run the command")

    if file_exts:
        for attachment in context.attachments:
            if not any(attachment.filename.endswith(ext) for ext in file_exts):
                raise errors.MissingRequiredAttachment("Missing attachment(s) required to run the command")

    return True


owner_only = Check(_owner_only)
"""Prevents a command from being used by anyone other than the owner of the application."""
guild_only = Check(_guild_only)
"""Prevents a command from being used in direct messages."""
dm_only = Check(_dm_only)
"""Prevents a command from being used in a guild."""
bot_only = Check(_bot_only)
"""Prevents a command from being used by anyone other than a bot."""
webhook_only = Check(_webhook_only)
"""Prevents a command from being used by anyone other than a webhook."""
human_only = Check(_human_only)
"""Prevents a command from being used by anyone other than a human."""
nsfw_channel_only = Check(_nsfw_channel_only)
"""
Prevents a command from being used in any channel other than one marked as NSFW.

.. deprecated:: 2.3.1
    Use the ``nsfw`` kwarg in the :obj:`~lightbulb.decorators.command` decorator instead. This will be removed
    in version ``2.4.0``.
"""


def has_roles(role1: int, *roles: int, mode: t.Callable[[t.Sequence[bool]], bool] = all) -> Check:
    """
    Prevents a command from being used by anyone missing roles according to the given mode.
    This check supports slash commands.

    Args:
        role1 (:obj:`int`): Role ID to check for.
        *roles (:obj:`int`): Additional role IDs to check for.

    Keyword Args:
        mode (``all`` or ``any``): The mode to check roles using. If ``all``, all role IDs passed will
            be required. If ``any``, only one of the role IDs will be required. Defaults to ``all``.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have roles
        in a DM channel.
    """
    if mode not in (any, all):
        raise TypeError("mode must be one of: any, all")
    return Check(functools.partial(_has_roles, roles=[role1, *roles], check_func=mode))


def has_guild_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used by a member missing any of the required
    permissions (this takes into account permissions granted by both roles and permission overwrites).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_has_guild_permissions, perms=reduced))


def has_role_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used by a member missing any of the required role
    permissions.

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_has_role_permissions, perms=reduced))


def has_channel_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used by a member missing any of the required
    channel permissions (permissions granted by a permission overwrite).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_has_channel_permissions, perms=reduced))


def bot_has_guild_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used if the bot is missing any of the required
    permissions (this takes into account permissions granted by both roles and permission overwrites).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_bot_has_guild_permissions, perms=reduced))


def bot_has_role_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used if the bot is missing any of the required role
    permissions.

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_bot_has_role_permissions, perms=reduced))


def bot_has_channel_permissions(perm1: hikari.Permissions, *perms: hikari.Permissions) -> Check:
    """
    Prevents the command from being used if the bot is missing any of the required channel
    permissions (permissions granted a permission overwrite).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *perms (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """
    reduced = functools.reduce(operator.or_, [perm1, *perms])
    return Check(functools.partial(_bot_has_channel_permissions, perms=reduced))


def has_attachments(*extensions: str) -> Check:
    """
    Prevents the command from being used if the invocation message
    does not include any attachments.

    Args:
        *extensions (:obj:`str`): If specified, attachments with different file extensions
            will cause the check to fail.

    Note:
        If ``extensions`` is specified then all attachments must conform to the restriction.
    """
    warnings.warn(
        "'has_attachments' is deprecated and scheduled for removal in version '2.5.0'. "
        "Use an option with type 'hikari.Attachment' instead.",
        DeprecationWarning,
    )
    return Check(functools.partial(_has_attachments, file_exts=extensions))



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["SlashCommand", "SlashCommandGroup", "SlashGroupMixin", "SlashSubGroup", "SlashSubCommand"]

import abc
import re
import typing as t

import hikari

from lightbulb import context as context_
from lightbulb import errors
from lightbulb.commands import base

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import plugins

COMMAND_NAME_REGEX: re.Pattern[str] = re.compile(r"^[\w-]{1,32}$", re.U)


class SlashGroupMixin(abc.ABC):
    __slots__ = ()
    _plugin: t.Optional[plugins.Plugin]
    _subcommands: t.Dict[str, t.Union[SlashSubGroup, SlashSubCommand]]

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    def create_subcommands(
        self,
        raw_cmds: t.Sequence[base.CommandLike],
        app: app_.BotApp,
        allowed_types: t.Union[t.Tuple[t.Type[SlashSubCommand], t.Type[SlashSubGroup]], t.Type[SlashSubCommand]],
    ) -> None:
        for raw_cmd in raw_cmds:
            impls: t.List[t.Type[base.Command]] = getattr(raw_cmd.callback, "__cmd_types__", [])
            for impl in impls:
                if issubclass(impl, allowed_types):
                    cmd = impl(app, raw_cmd)
                    assert isinstance(cmd, (SlashSubCommand, SlashSubGroup))
                    cmd.parent = self  # type: ignore
                    cmd.plugin = self.plugin  # type: ignore

                    if cmd.name in self._subcommands:
                        raise errors.CommandAlreadyExists(
                            f"A prefix subcommand with name or alias {cmd.name!r} already exists for group {self.name!r}"
                        )
                    self._subcommands[cmd.name] = cmd

    def recreate_subcommands(self, raw_cmds: t.Sequence[base.CommandLike], app: app_.BotApp) -> None:
        self._subcommands.clear()
        self.create_subcommands(
            raw_cmds, app, SlashSubCommand if isinstance(self, SlashSubGroup) else (SlashSubCommand, SlashSubGroup)
        )

    async def _invoke_subcommand(self, context: context_.base.Context) -> None:
        assert isinstance(context, context_.slash.SlashContext)
        cmd_option = context._raw_options[0]
        context._raw_options = cmd_option.options or []
        # Replace the invoked command prematurely so that _parse_options uses the correct command options
        context._invoked = self._subcommands[cmd_option.name]
        # Reparse the options for the subcommand
        context._parse_options(cmd_option.options)
        # Ensure we call _maybe_defer
        await context._maybe_defer()
        # Invoke the subcommand
        await context._invoked.invoke(context)

    def get_subcommand(self, name: str) -> t.Optional[t.Union[SlashSubGroup, SlashSubCommand]]:
        """Get the group's subcommand with the given name."""
        return self._subcommands.get(name)

    @property
    def subcommands(self) -> t.Dict[str, t.Union[SlashSubGroup, SlashSubCommand]]:
        """Mapping of command name to command object containing the group's subcommands."""
        return self._subcommands

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        self._plugin = pl
        for command in self._subcommands.values():
            if isinstance(command, SlashGroupMixin):
                command._set_plugin(pl)
            else:
                command.plugin = pl


class SlashCommand(base.ApplicationCommand):
    """
    An implementation of :obj:`~.commands.base.Command` representing a slash command.

    See the `API Documentation <https://discord.com/developers/docs/interactions/application-commands#slash-commands>`_.
    """

    __slots__ = ()

    def as_create_kwargs(self) -> t.Dict[str, t.Any]:
        sorted_opts = sorted(self.options.values(), key=lambda o: int(o.required), reverse=True)
        return {
            "type": hikari.CommandType.SLASH,
            "name": self.name,
            "description": self.description,
            "options": [o.as_application_command_option() for o in sorted_opts],
        }

    def _validate_attributes(self) -> None:
        if not COMMAND_NAME_REGEX.fullmatch(self.name) or self.name != self.name.lower():
            raise ValueError(
                f"Slash command {self.name!r}: name must match regex '^[\\w-]{1,32}$' and be all lowercase"
            ) from None
        if len(self.description) < 1 or len(self.description) > 100:
            raise ValueError(f"Slash command {self.name!r}: description must be from 1-100 characters long") from None
        if len(self.options) > 25:
            raise ValueError(f"Slash command {self.name!r}: can at most have 25 options") from None


class SlashSubCommand(SlashCommand, base.SubCommandTrait):
    """
    Class representing a slash subcommand.
    """

    __slots__ = ()

    @property
    def qualname(self) -> str:
        assert self.parent is not None
        return f"{self.parent.qualname} {self.name}"

    def as_option(self) -> hikari.CommandOption:
        sorted_opts = sorted(self.options.values(), key=lambda o: int(o.required), reverse=True)
        return hikari.CommandOption(
            type=hikari.OptionType.SUB_COMMAND,
            name=self.name,
            description=self.description,
            is_required=False,
            options=[o.as_application_command_option() for o in sorted_opts],
        )


class SlashSubGroup(SlashCommand, SlashGroupMixin, base.SubCommandTrait):
    """
    Class representing a slash subgroup of commands.
    """

    __slots__ = ("_raw_subcommands", "_subcommands")

    def __init__(self, app: app_.BotApp, initialiser: base.CommandLike) -> None:
        super().__init__(app, initialiser)
        self._raw_subcommands = initialiser.subcommands
        initialiser.subcommands = (
            initialiser.subcommands.add_parent(self)  # type: ignore
            if isinstance(initialiser.subcommands, base._SubcommandListProxy)  # type: ignore
            else base._SubcommandListProxy(initialiser.subcommands, parent=self)
        )
        # Just to keep mypy happy we leave SlashSubGroup here
        self._subcommands: t.Dict[str, t.Union[SlashSubGroup, SlashSubCommand]] = {}
        self.create_subcommands(self._raw_subcommands, app, SlashSubCommand)

    @property
    def qualname(self) -> str:
        assert self.parent is not None
        return f"{self.parent.qualname} {self.name}"

    def as_option(self) -> hikari.CommandOption:
        return hikari.CommandOption(
            type=hikari.OptionType.SUB_COMMAND_GROUP,
            name=self.name,
            description=self.description,
            is_required=False,
            options=[c.as_option() for c in self._subcommands.values()],
        )

    async def invoke(self, context: context_.base.Context, **_: t.Any) -> None:
        await self._invoke_subcommand(context)

    def _validate_attributes(self) -> None:
        super()._validate_attributes()
        if len(self._subcommands) > 25:
            raise ValueError(f"Slash command {self.name!r}: group can have at most 25 subcommands") from None

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        SlashGroupMixin._set_plugin(self, pl)


class SlashCommandGroup(SlashCommand, SlashGroupMixin):
    """
    Class representing a slash command group.
    """

    __slots__ = ("_raw_subcommands", "_subcommands")

    def __init__(self, app: app_.BotApp, initialiser: base.CommandLike) -> None:
        super().__init__(app, initialiser)
        self._raw_subcommands = initialiser.subcommands
        initialiser.subcommands = (
            initialiser.subcommands.add_parent(self)  # type: ignore
            if isinstance(initialiser.subcommands, base._SubcommandListProxy)  # type: ignore
            else base._SubcommandListProxy(initialiser.subcommands, parent=self)
        )
        self._subcommands: t.Dict[str, t.Union[SlashSubGroup, SlashSubCommand]] = {}
        self.create_subcommands(self._raw_subcommands, app, (SlashSubCommand, SlashSubGroup))

    async def invoke(self, context: context_.base.Context, **_: t.Any) -> None:
        await self._invoke_subcommand(context)

    def as_create_kwargs(self) -> t.Dict[str, t.Any]:
        return {
            "type": hikari.CommandType.SLASH,
            "name": self.name,
            "description": self.description,
            "options": [c.as_option() for c in self._subcommands.values()],
        }

    def _validate_attributes(self) -> None:
        super()._validate_attributes()
        if len(self._subcommands) > 25:
            raise ValueError(f"Slash command {self.name!r}: group can have at most 25 subcommands") from None

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        SlashGroupMixin._set_plugin(self, pl)



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["UserCommand"]

import typing as t

import hikari

from lightbulb.commands import base


class UserCommand(base.ApplicationCommand):
    """
    An implementation of :obj:`~.commands.base.Command` representing a user context menu command.

    See the `API Documentation <https://discord.com/developers/docs/interactions/application-commands#user-commands>`_.

    .. versionadded:: 2.2.0
    """

    __slots__ = ()

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

    @property
    def options(self) -> t.Dict[str, t.Any]:
        return {}

    @property
    def description_localizations(self) -> t.Mapping[t.Union[hikari.Locale, str], str]:
        return {}

    def _validate_attributes(self) -> None:
        if len(self.name) < 1 or len(self.name) > 32:
            raise ValueError(f"User command {self.name!r}: name must be from 1-32 characters long") from None

    def as_create_kwargs(self) -> t.Dict[str, t.Any]:
        return {
            "type": hikari.CommandType.USER,
            "name": self.name,
        }



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "ReactionNavigator",
    "ButtonNavigator",
    "ReactionButton",
    "ComponentButton",
    "next_page",
    "prev_page",
    "first_page",
    "last_page",
    "stop",
]

import asyncio
import typing as t

import hikari
from hikari.api.special_endpoints import MessageActionRowBuilder
from hikari.components import ButtonStyle

from lightbulb import context as context_


class ReactionButton:
    """
    A reaction-based navigator button. Contains the emoji linked to the button as well as
    the coroutine to be called when the button is pressed.

    Args:
        emoji (Union[:obj:`str`, :obj:`~hikari.emojis.Emoji`]): The emoji linked to the button.
        callback: The coroutine function to be called on button press.
    """

    __slots__ = ("emoji", "callback")

    def __init__(
        self,
        emoji: t.Union[str, hikari.Emoji],
        callback: t.Callable[[ReactionNavigator[T], hikari.ReactionAddEvent], t.Coroutine[t.Any, t.Any, None]],
    ) -> None:
        if isinstance(emoji, str):
            emoji = hikari.Emoji.parse(emoji)
        self.emoji = emoji
        self.callback = callback

    def is_pressed(self, event: hikari.ReactionAddEvent) -> bool:
        """
        Check if the button is pressed in a given event.

        Args:
            event (:obj:`~hikari.events.message.MessageReactionEvent`): The event to check the button is pressed in.

        Returns:
            :obj:`bool`: Whether the button is pressed in the given event.
        """
        if event.emoji_id is not None and isinstance(self.emoji, hikari.CustomEmoji):
            return event.emoji_id == self.emoji.id
        return event.emoji_name == self.emoji.name

    def press(self, nav: ReactionNavigator[T], event: hikari.ReactionAddEvent) -> t.Coroutine[t.Any, t.Any, None]:
        """
        Call the button's callback coroutine and return the awaitable.

        Returns:
            Coroutine[``None``, Any, ``None``]: Returned awaitable from the coroutine call.
        """
        return self.callback(nav, event)


class ComponentButton:
    """
    A component-based navigator button. Contains the custom_id linked to the button as well as
    the coroutine to be called when the button is pressed.

    Args:
        label (:obj:`str`): The label of the button.
        label_is_emoji (:obj:`bool`): Whether the label is an emoji or not. This affects whether ``set_label`` or
            ``set_emoji`` is called when building the button.
        style (:obj:`hikari.InteractiveButtonTypesT`): The style of the button.
        custom_id (:obj:`str`): The custom ID of the button.
        callback: The coroutine function to be called on button press.
    """

    __slots__ = ("custom_id", "callback", "label", "style", "label_is_emoji")

    def __init__(
        self,
        label: str,
        label_is_emoji: bool,
        style: hikari.InteractiveButtonTypesT,
        custom_id: str,
        callback: t.Callable[[ButtonNavigator[T], hikari.InteractionCreateEvent], t.Coroutine[t.Any, t.Any, None]],
    ) -> None:
        self.label = label
        self.label_is_emoji = label_is_emoji
        self.style = style
        self.custom_id = custom_id
        self.callback = callback

    def build(self, container: MessageActionRowBuilder, disabled: bool = False) -> None:
        """
        Build and add the button to the given container.

        Args:
            container (:obj:`hikari.api.special_endpoints.MessageActionRowBuilder`): The container to add the button to.
            disabled (:obj:`bool`): Whether to display the button as disabled.

        Returns:
            ``None``
        """
        container.add_interactive_button(
            self.style,
            self.custom_id,
            is_disabled=disabled,
            **{"emoji" if self.label_is_emoji else "label": self.label},
        )

    def is_pressed(self, event: hikari.InteractionCreateEvent) -> bool:
        """
        Check if the button is pressed in a given event.

        Args:
            event (:obj:`~hikari.InteractionCreateEvent`): The event to check the button is pressed in.

        Returns:
            :obj:`bool`: Whether or not the button is pressed in the given event.
        """
        assert isinstance(event.interaction, hikari.ComponentInteraction)
        return event.interaction.custom_id == self.custom_id

    def press(self, nav: ButtonNavigator[T], event: hikari.InteractionCreateEvent) -> t.Coroutine[t.Any, t.Any, None]:
        """
        Call the button's callback coroutine and return the awaitable.

        Returns:
            Coroutine[Any, Any, ``None``]: Returned awaitable from the coroutine call.
        """
        return self.callback(nav, event)


T = t.TypeVar("T")


async def next_page(nav: t.Union[ReactionNavigator[T], ButtonNavigator[T]], _: hikari.Event) -> None:
    """
    :obj:`NavButton` callback to make the ReactionNavigator go to the next page.
    """
    nav.current_page_index += 1
    nav.current_page_index %= len(nav.pages)


async def prev_page(nav: t.Union[ReactionNavigator[T], ButtonNavigator[T]], _: hikari.Event) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the previous page.
    """
    nav.current_page_index -= 1
    if nav.current_page_index < 0:
        nav.current_page_index = len(nav.pages) - 1


async def first_page(nav: t.Union[ReactionNavigator[T], ButtonNavigator[T]], _: hikari.Event) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the first page.
    """
    nav.current_page_index = 0


async def last_page(nav: t.Union[ReactionNavigator[T], ButtonNavigator[T]], _: hikari.Event) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the last page.
    """
    nav.current_page_index = len(nav.pages) - 1


async def stop(nav: t.Union[ReactionNavigator[T], ButtonNavigator[T]], _: hikari.Event) -> None:
    """
    :obj:`NavButton` callback to make the navigator stop navigation.
    """
    assert nav._msg is not None
    await nav._remove_listener()
    await nav._msg.delete()
    nav._msg = None
    if nav._timeout_task is not None:
        nav._timeout_task.cancel()


class ReactionNavigator(t.Generic[T]):
    """
    A reaction navigator system for navigating through a list of items that can be sent through the
    ``content`` argument of :obj:`hikari.Message.respond`.

    Default buttons:

    - ``\\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\\N{VARIATION SELECTOR-16}`` (Go to first page)
    - ``\\N{BLACK LEFT-POINTING TRIANGLE}\\N{VARIATION SELECTOR-16}`` (Go to previous page)
    - ``\\N{BLACK SQUARE FOR STOP}\\N{VARIATION SELECTOR-16}`` (Stop navigation)
    - ``\\N{BLACK RIGHT-POINTING TRIANGLE}\\N{VARIATION SELECTOR-16}`` (Go to next page)
    - ``\\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\\N{VARIATION SELECTOR-16}`` (Go to last page)

    Args:
        pages (Sequence[T]): Pages to navigate through.

    Keyword Args:
        buttons (Optional[Sequence[:obj:`~.utils.nav.NavButton`]]): Buttons to
            use the navigator with. Uses the default buttons if not specified.
        timeout (:obj:`float`): The navigator timeout in seconds. After the timeout has expired, navigator reactions
            will no longer work. Defaults to 120 (2 minutes).

    Example:
        .. code-block:: python

            from lightbulb.utils import pag, nav

            @bot.command()
            async def foo(ctx):
                paginated_help = pag.StringPaginator()
                for l in thing_that_creates_a_lot_of_text.split("\\n"):
                    paginated_help.add_line(l)
                navigator = nav.ReactionNavigator(paginated_help.build_pages())
                await navigator.run(ctx)
    """

    def __init__(
        self,
        pages: t.Union[t.Iterable[T], t.Iterator[T]],
        *,
        buttons: t.Optional[t.Sequence[ReactionButton]] = None,
        timeout: float = 120,
    ) -> None:
        if not pages:
            raise ValueError("You cannot pass fewer than 1 page to the navigator.")
        self.pages: t.Sequence[T] = tuple(pages)

        if buttons is not None:
            if any(not isinstance(btn, ReactionButton) for btn in buttons):
                raise TypeError("Buttons must be an instance of ReactionButton")

        self.buttons: t.Sequence[ReactionButton]
        if len(self.pages) == 1 and not buttons:
            self.buttons = [ReactionButton("\N{CROSS MARK}", stop)]
        else:
            self.buttons = buttons if buttons is not None else self.create_default_buttons()

        self._timeout: float = timeout
        self.current_page_index: int = 0
        self._context: t.Optional[context_.base.Context] = None
        self._msg: t.Optional[hikari.Message] = None
        self._timeout_task: t.Optional[asyncio.Task[None]] = None

    async def _edit_msg(self, message: hikari.Message, page: T) -> hikari.Message:
        return await message.edit(page)

    async def _send_initial_msg(self, page: T) -> hikari.Message:
        assert self._context is not None
        resp = await self._context.respond(page)
        return await resp.message()

    def create_default_buttons(self) -> t.List[ReactionButton]:
        buttons = [
            ReactionButton(
                "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}", first_page
            ),
            ReactionButton("\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}", prev_page),
            ReactionButton("\N{BLACK SQUARE FOR STOP}\N{VARIATION SELECTOR-16}", stop),
            ReactionButton("\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}", next_page),
            ReactionButton(
                "\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}", last_page
            ),
        ]
        return buttons

    async def _process_reaction_add(self, event: hikari.ReactionAddEvent) -> None:
        assert self._context is not None and self._msg is not None
        if event.user_id != self._context.author.id or event.message_id != self._msg.id:
            return

        for button in self.buttons:
            if button.is_pressed(event):
                await button.press(self, event)
                if self._msg is not None:
                    await self._edit_msg(self._msg, self.pages[self.current_page_index])
                    try:
                        await self._msg.remove_reaction(button.emoji, user=self._context.author)
                    except hikari.ForbiddenError:
                        pass
                break

    async def _remove_listener(self) -> None:
        assert self._context is not None
        self._context.app.unsubscribe(hikari.ReactionAddEvent, self._process_reaction_add)

        if self._msg is None:
            return

        try:
            await self._msg.remove_all_reactions()
        except (hikari.ForbiddenError, hikari.NotFoundError):
            pass

    async def _timeout_coro(self) -> None:
        try:
            await asyncio.sleep(self._timeout)
            await self._remove_listener()
        except asyncio.CancelledError:
            pass

    async def run(self, context: context_.base.Context) -> None:
        """
        Run the navigator under the given context.

        Args:
            context (:obj:`~.context.base.Context`): Context
                to run the navigator under.

        Returns:
            ``None``

        Raises:
            :obj:`hikari.MissingIntentError`: If the bot does not have the relevant reaction intent(s) for
                the navigator to function.
        """
        intent_to_check_for = (
            hikari.Intents.GUILD_MESSAGE_REACTIONS
            if context.guild_id is not None
            else hikari.Intents.DM_MESSAGE_REACTIONS
        )
        if not (context.app.intents & intent_to_check_for) == intent_to_check_for:
            raise hikari.MissingIntentError(intent_to_check_for)

        self._context = context
        context.app.subscribe(hikari.ReactionAddEvent, self._process_reaction_add)
        self._msg = await self._send_initial_msg(self.pages[self.current_page_index])
        for emoji in [button.emoji for button in self.buttons]:
            await self._msg.add_reaction(emoji)

        if self._timeout_task is not None:
            self._timeout_task.cancel()
        self._timeout_task = asyncio.create_task(self._timeout_coro())


class ButtonNavigator(t.Generic[T]):
    """
    A button navigator system for navigating through a list of items that can be sent through the
    ``content`` argument of :obj:`hikari.Message.respond`.

    Args:
        pages (Sequence[T]): Pages to navigate through.

    Keyword Args:
        buttons (Sequence[:obj:`~ComponentButton`]): Buttons to
            use the navigator with. Uses the default buttons if not specified.
        timeout (:obj:`float`): The navigator timeout in seconds. After the timeout has expired, navigator buttons
            are disabled and will no longer work. Defaults to 120 (2 minutes).

    Example:
        .. code-block:: python

            from lightbulb.utils import pag, nav

            @bot.command()
            async def foo(ctx):
                paginated_help = pag.StringPaginator()
                for l in thing_that_creates_a_lot_of_text.split("\\n"):
                    paginated_help.add_line(l)
                navigator = nav.ButtonNavigator(paginated_help.build_pages())
                await navigator.run(ctx)
    """

    def __init__(
        self,
        pages: t.Union[t.Iterable[T], t.Iterator[T]],
        *,
        buttons: t.Optional[t.Sequence[ComponentButton]] = None,
        timeout: float = 120,
    ) -> None:
        if not pages:
            raise ValueError("You cannot pass fewer than 1 page to the navigator.")
        self.pages: t.Sequence[T] = tuple(pages)

        self.buttons: t.Sequence[ComponentButton]
        if len(self.pages) == 1 and not buttons:
            self.buttons = [ComponentButton("\N{CROSS MARK}", True, ButtonStyle.DANGER, "stop", stop)]
        else:
            self.buttons = buttons if buttons is not None else self.create_default_buttons()

        self._timeout: float = timeout
        self.current_page_index: int = 0
        self._context: t.Optional[context_.base.Context] = None
        self._msg: t.Optional[hikari.Message] = None
        self._timeout_task: t.Optional[asyncio.Task[None]] = None

    async def _send_initial_msg(self, page: T) -> hikari.Message:
        assert self._context is not None
        buttons = await self.build_buttons()
        resp = await self._context.respond(page, component=buttons)
        return await resp.message()

    async def _edit_msg(self, inter: hikari.ComponentInteraction, page: T) -> None:
        buttons = await self.build_buttons(disabled=True if self._msg is None else False)
        try:
            await inter.create_initial_response(hikari.ResponseType.MESSAGE_UPDATE, page, component=buttons)
        except hikari.NotFoundError:
            await inter.edit_initial_response(page, component=buttons)

    def create_default_buttons(self) -> t.Sequence[ComponentButton]:
        buttons = [
            ComponentButton(
                "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}",
                True,
                ButtonStyle.PRIMARY,
                "first_page",
                first_page,
            ),
            ComponentButton(
                "\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
                True,
                ButtonStyle.PRIMARY,
                "prev_page",
                prev_page,
            ),
            ComponentButton(
                "\N{HEAVY MULTIPLICATION X}\N{VARIATION SELECTOR-16}", True, ButtonStyle.DANGER, "stop", stop
            ),
            ComponentButton(
                "\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
                True,
                ButtonStyle.PRIMARY,
                "next_page",
                next_page,
            ),
            ComponentButton(
                "\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}",
                True,
                ButtonStyle.PRIMARY,
                "last_page",
                last_page,
            ),
        ]
        return buttons

    async def build_buttons(self, disabled: bool = False) -> hikari.UndefinedOr[MessageActionRowBuilder]:
        assert self._context is not None
        buttons = self._context.app.rest.build_message_action_row()
        for button in self.buttons:
            button.build(buttons, disabled)
        return buttons

    async def _process_interaction_create(self, event: hikari.InteractionCreateEvent) -> None:
        if not isinstance(event.interaction, hikari.ComponentInteraction):
            return

        if self._msg is None:
            return

        assert self._context is not None

        if event.interaction.message.id != self._msg.id or event.interaction.user.id != self._context.author.id:
            return

        for button in self.buttons:
            if button.is_pressed(event):
                await button.press(self, event)
                if self._msg is not None:
                    await self._edit_msg(event.interaction, self.pages[self.current_page_index])
                break

    async def _remove_listener(self) -> None:
        assert self._context is not None
        self._context.app.unsubscribe(hikari.InteractionCreateEvent, self._process_interaction_create)

        if self._msg is not None:
            await self._msg.edit(component=await self.build_buttons(True))

    async def _timeout_coro(self) -> None:
        try:
            await asyncio.sleep(self._timeout)
            await self._remove_listener()
        except asyncio.CancelledError:
            pass

    async def run(self, context: context_.base.Context) -> None:
        """
        Run the navigator under the given context.

        Args:
            context (:obj:`~.context.base.Context`): Context
                to run the navigator under.

        Returns:
            ``None``
        """
        self._context = context
        context.app.subscribe(hikari.InteractionCreateEvent, self._process_interaction_create)
        self._msg = await self._send_initial_msg(self.pages[self.current_page_index])

        if self._timeout_task is not None:
            self._timeout_task.cancel()
        self._timeout_task = asyncio.create_task(self._timeout_coro())



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["permissions_for", "permissions_in"]

import hikari


def permissions_for(member: hikari.Member) -> hikari.Permissions:
    """
    Get the guild permissions for the given member.

    Args:
        member (:obj:`hikari.Member`): Member to get permissions for.

    Returns:
        :obj:`hikari.Permissions`: Member's guild permissions.

    Warning:
        This method relies on the cache to work. If the cache is not available then :obj:`hikari.Permissions.NONE`
        will be returned.
    """
    permissions = hikari.Permissions.NONE
    for role in member.get_roles():
        permissions |= role.permissions

    guild = member.get_guild()

    if hikari.Permissions.ADMINISTRATOR in permissions or guild and member.id == guild.owner_id:
        return hikari.Permissions.all_permissions()

    return permissions


def permissions_in(
    channel: hikari.PermissibleGuildChannel, member: hikari.Member, include_guild_permissions: bool = True
) -> hikari.Permissions:
    """
    Get the permissions for the given member in the given guild channel.

    Args:
        channel (:obj:`hikari.GuildChannel`): Channel to get the permissions in.
        member (:obj:`hikari.Member`): Member to get the permissions for.
        include_guild_permissions (:obj:`bool`): Whether or not to include the member's guild permissions. If ``False``,
            only permissions granted by overwrites will be included. Defaults to ``True``.

    Returns:
        :obj:`hikari.Permissions`: Member's permissions in the given channel.
    """
    allowed_perms = hikari.Permissions.NONE
    if include_guild_permissions:
        allowed_perms |= permissions_for(member)

    guild = member.get_guild()

    if hikari.Permissions.ADMINISTRATOR in allowed_perms or guild and member.id == guild.owner_id:
        return hikari.Permissions.all_permissions()

    overwrites = channel.permission_overwrites

    permissions = allowed_perms
    if overwrite_everyone := overwrites.get(member.guild_id):
        permissions &= ~overwrite_everyone.deny
        permissions |= overwrite_everyone.allow

    allow = hikari.Permissions.NONE
    deny = hikari.Permissions.NONE
    for role_id in member.role_ids:
        if overwrite_role := overwrites.get(role_id):
            allow |= overwrite_role.allow
            deny |= overwrite_role.deny

    permissions &= ~deny
    permissions |= allow

    if overwrite_member := overwrites.get(member.id):
        permissions &= ~overwrite_member.deny
        permissions |= overwrite_member.allow

    return permissions



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["serialise_command", "manage_application_commands"]

import logging
import typing as t

import hikari

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb.commands import base

_LOGGER = logging.getLogger("lightbulb.internal")


class _GuildIDCollection:
    __slots__ = ("ids",)

    def __init__(self, ids: t.Optional[t.Union[t.Sequence[int], int]]):
        if ids is None:
            ids = []
        elif isinstance(ids, int):
            ids = [ids]
        self.ids: t.List[int] = list(ids)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, int):
            return NotImplemented
        return other in self.ids

    def __repr__(self) -> str:
        return f"Guilds({', '.join(str(i) for i in self.ids)})"


def _serialise_option(option: hikari.CommandOption) -> t.Dict[str, t.Any]:
    return {
        "type": option.type,
        "name": option.name,
        "description": option.description,
        "is_required": option.is_required,
        "choices": list(
            sorted(
                [{"n": c.name, "v": c.value} for c in option.choices] if option.choices is not None else [],
                key=lambda d: d["n"],  # type: ignore
            )
        ),
        "options": [_serialise_option(o) for o in option.options] if option.options is not None else [],
        "channel_types": list(sorted(option.channel_types if option.channel_types is not None else [])),
        "min_value": option.min_value,
        "max_value": option.max_value,
        "min_length": option.min_length,
        "max_length": option.max_length,
        "autocomplete": option.autocomplete,
    }


def _serialise_hikari_command(command: hikari.PartialCommand) -> t.Dict[str, t.Any]:
    options = sorted(getattr(command, "options", []) or [], key=lambda o: o.name)
    return {
        "type": command.type,
        "name": command.name,
        "description": getattr(command, "description", None) or None,
        "options": [_serialise_option(o) for o in options],
        "guild_id": command.guild_id or None,
        "default_member_permissions": command.default_member_permissions or None,
        "dm_enabled": command.is_dm_enabled,
        "nsfw": command.is_nsfw,
    }


def _serialise_lightbulb_command(command: base.ApplicationCommand) -> t.Dict[str, t.Any]:
    create_kwargs = command.as_create_kwargs()
    return {
        "type": create_kwargs["type"],
        "name": create_kwargs["name"],
        "description": create_kwargs.get("description"),
        "options": [_serialise_option(o) for o in sorted(create_kwargs.get("options", []), key=lambda o: o.name)],  # type: ignore
        "guild_id": _GuildIDCollection(command.guilds) if command.guilds else None,
        "default_member_permissions": command.app_command_default_member_permissions,
        "dm_enabled": command.app_command_dm_enabled if not command.guilds else False,
        "nsfw": command.nsfw,
    }


def serialise_command(command: t.Union[hikari.PartialCommand, base.ApplicationCommand]) -> t.Dict[str, t.Any]:
    if isinstance(command, hikari.PartialCommand):
        return _serialise_hikari_command(command)
    return _serialise_lightbulb_command(command)


def _compare_commands(cmd1: base.ApplicationCommand, cmd2: hikari.PartialCommand) -> bool:
    return serialise_command(cmd1) == serialise_command(cmd2)


def _create_builder_from_command(
    app: app_.BotApp, cmd: t.Union[hikari.PartialCommand, base.ApplicationCommand]
) -> t.Union[hikari.api.SlashCommandBuilder, hikari.api.ContextMenuCommandBuilder]:
    bld: t.Union[hikari.api.SlashCommandBuilder, hikari.api.ContextMenuCommandBuilder]
    if isinstance(cmd, hikari.PartialCommand):
        if desc := getattr(cmd, "description", None):
            bld = app.rest.slash_command_builder(cmd.name, description=desc)
            for option in getattr(cmd, "options", []) or []:
                bld.add_option(option)
        else:
            bld = app.rest.context_menu_command_builder(cmd.type, cmd.name)

        bld.set_default_member_permissions(cmd.default_member_permissions)
        bld.set_is_dm_enabled(cmd.is_dm_enabled)
        bld.set_is_nsfw(cmd.is_nsfw)

        bld.set_id(cmd.id)
    else:
        create_kwargs = cmd.as_create_kwargs()
        if "description" in create_kwargs:
            bld = app.rest.slash_command_builder(create_kwargs["name"], description=create_kwargs["description"])
            for opt in create_kwargs.get("options", []):
                bld.add_option(opt)
        else:
            bld = app.rest.context_menu_command_builder(create_kwargs["type"], create_kwargs["name"])

        if cmd.app_command_default_member_permissions is not None:
            bld.set_default_member_permissions(cmd.app_command_default_member_permissions)
        bld.set_is_dm_enabled(cmd.app_command_dm_enabled)
        bld.set_is_nsfw(cmd.nsfw)

    return bld


def _get_lightbulb_command_equivalent(
    app: app_.BotApp, cmd: hikari.PartialCommand
) -> t.Optional[base.ApplicationCommand]:
    if cmd.type is hikari.CommandType.SLASH:
        return app.get_slash_command(cmd.name)
    elif cmd.type is hikari.CommandType.USER:
        return app.get_user_command(cmd.name)
    elif cmd.type is hikari.CommandType.MESSAGE:
        return app.get_message_command(cmd.name)


async def _get_guild_commands_to_set(app: app_.BotApp, guild_id: int) -> t.Sequence[hikari.api.CommandBuilder]:
    assert app.application is not None

    commands_to_declare = []
    unchanged, changed, created, deleted, skipped = 0, 0, 0, 0, 0
    # Get the commands that already exist in the given guild
    existing_commands = await app.rest.fetch_application_commands(app.application, guild_id)

    # Create a mapping containing all the commands that should be created for this guild
    app_commands: t.Dict[hikari.CommandType, t.Dict[str, base.ApplicationCommand]] = {
        hikari.CommandType.SLASH: {c.name: c for c in app._slash_commands.values() if guild_id in c.guilds},
        hikari.CommandType.USER: {c.name: c for c in app._user_commands.values() if guild_id in c.guilds},
        hikari.CommandType.MESSAGE: {c.name: c for c in app._message_commands.values() if guild_id in c.guilds},
    }
    for command in existing_commands:
        # Get the implementation for the given command
        equiv = _get_lightbulb_command_equivalent(app, command)
        if equiv is not None and guild_id not in equiv.guilds:
            equiv = None

        if equiv is None and not app._delete_unbound_commands:
            # Convert the hikari Command back into a builder because we don't want to delete it
            commands_to_declare.append(_create_builder_from_command(app, command))
            skipped += 1
        elif equiv is None and app._delete_unbound_commands:
            deleted += 1
        elif equiv is not None:
            if _compare_commands(equiv, command):
                # The commands are the same, create a builder from the hikari Command
                commands_to_declare.append(_create_builder_from_command(app, command))
                unchanged += 1
            else:
                # The commands are different, create a builder from the lightbulb Command
                commands_to_declare.append(_create_builder_from_command(app, equiv))
                changed += 1
        # Remove the command from the mapping, so we don't add a duplicate to the list
        app_commands[command.type].pop(command.name, None)

    # Add the remaining commands that do not have an equivalent on discord
    for cmds in app_commands.values():
        for cmd in cmds.values():
            commands_to_declare.append(_create_builder_from_command(app, cmd))
            created += 1

    _LOGGER.info("Processing application commands for guild %s", guild_id)
    _LOGGER.debug(
        "%s - Created: %s, Deleted: %s, Updated: %s, Unchanged: %s, Skipped: %s",
        guild_id,
        created,
        deleted,
        changed,
        unchanged,
        skipped,
    )

    return commands_to_declare


async def _process_global_commands(app: app_.BotApp) -> None:
    assert app.application is not None

    unchanged, changed, created, deleted, skipped = 0, 0, 0, 0, 0
    # Get the commands that already exist globally
    existing_global_commands = await app.rest.fetch_application_commands(app.application)
    # Create a mapping containing all the commands that should be created globally
    registered_global_commands: t.Dict[hikari.CommandType, t.Dict[str, base.ApplicationCommand]] = {
        hikari.CommandType.SLASH: {c.name: c for c in app._slash_commands.values() if not c.guilds},
        hikari.CommandType.USER: {c.name: c for c in app._user_commands.values() if not c.guilds},
        hikari.CommandType.MESSAGE: {c.name: c for c in app._message_commands.values() if not c.guilds},
    }
    for command in existing_global_commands:
        # Get the implementation for the given command
        equiv = registered_global_commands[command.type].get(command.name)
        if equiv is None:
            # No implementation for the hikari Command exists as a global command
            if app._delete_unbound_commands:
                # Delete the command
                await command.delete()
                deleted += 1
            else:
                # Ignore the command
                skipped += 1
        else:
            if _compare_commands(equiv, command):
                # The commands are the same, no need to recreate
                unchanged += 1
                equiv.instances[None] = command
            else:
                # The commands are different, recreate in order to update the version on discord
                await equiv.create()
                changed += 1
        # Remove the command from the mapping, so we don't make too many create calls
        registered_global_commands[command.type].pop(command.name, None)

    # Create the remaining commands that do not have an equivalent on discord
    for cmd_map in registered_global_commands.values():
        for app_cmd in cmd_map.values():
            await app_cmd.create()
            created += 1

    _LOGGER.debug(
        "Global - Created: %s, Deleted: %s, Updated: %s, Unchanged: %s, Skipped: %s",
        created,
        deleted,
        changed,
        unchanged,
        skipped,
    )


async def manage_application_commands(app: app_.BotApp) -> None:
    assert app.application is not None

    # Guild command processing
    _LOGGER.info("Processing guild application commands")
    all_guilds = set(app.default_enabled_guilds)
    for app_cmd in [*app._message_commands.values(), *app._user_commands.values(), *app._slash_commands.values()]:
        all_guilds.update(app_cmd.guilds or [])

    cmd_mapping: t.Dict[hikari.CommandType, t.Dict[str, base.ApplicationCommand]] = {
        hikari.CommandType.SLASH: app._slash_commands,  # type: ignore[dict-item]
        hikari.CommandType.USER: app._user_commands,  # type: ignore[dict-item]
        hikari.CommandType.MESSAGE: app._message_commands,  # type: ignore[dict-item]
    }
    for guild_id in all_guilds:
        cmds_to_declare = await _get_guild_commands_to_set(app, guild_id)
        created = await app.rest.set_application_commands(app.application, cmds_to_declare, guild_id)
        for created_cmd in created:
            if equiv := cmd_mapping[created_cmd.type].get(created_cmd.name):
                equiv.instances[guild_id] = created_cmd

    # Global command processing
    _LOGGER.info("Processing global application commands")
    await _process_global_commands(app)

    _LOGGER.info("Application command processing completed")



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["BotApp", "when_mentioned_or"]

import asyncio
import collections.abc
import functools
import importlib
import inspect
import logging
import os
import pathlib
import re
import sys
import typing as t
from importlib import util

import hikari
from multidict import CIMultiDict

from lightbulb import checks
from lightbulb import commands
from lightbulb import context as context_
from lightbulb import decorators
from lightbulb import errors
from lightbulb import events
from lightbulb import help_command as help_command_
from lightbulb import internal
from lightbulb import parser
from lightbulb import plugins as plugins_
from lightbulb.utils import data_store

_LOGGER = logging.getLogger("lightbulb.app")
_APPLICATION_CMD_ERROR_REGEX: re.Pattern[str] = re.compile(
    r"https?://discord.com/api/v\d+/applications/\d+/guilds/(\d+)/commands"
)

PrefixT = t.Union[
    t.Sequence[str],
    t.Callable[["BotApp", hikari.Message], t.Union[t.Sequence[str], t.Coroutine[t.Any, t.Any, t.Sequence[str]]]],
]
CheckCoroT = t.TypeVar("CheckCoroT", bound=t.Callable[..., t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]])


class _ExtensionT(t.Protocol):
    def load(self, bot: BotApp) -> None:
        ...

    def unload(self, bot: BotApp) -> None:
        ...


APPLICATION_COMMANDS_EVENTS_MAPPING = {
    commands.slash.SlashCommand: (
        events.SlashCommandInvocationEvent,
        events.SlashCommandCompletionEvent,
        events.SlashCommandErrorEvent,
    ),
    commands.message.MessageCommand: (
        events.MessageCommandInvocationEvent,
        events.MessageCommandCompletionEvent,
        events.MessageCommandErrorEvent,
    ),
    commands.user.UserCommand: (
        events.UserCommandInvocationEvent,
        events.UserCommandCompletionEvent,
        events.UserCommandErrorEvent,
    ),
}


def when_mentioned_or(
    prefix_provider: PrefixT,
) -> t.Callable[[BotApp, hikari.Message], t.Coroutine[t.Any, t.Any, t.Sequence[str]]]:
    """
    Helper function which allows the bot's mentions to be used as the command prefix, as well
    as any other prefix(es) passed in or supplied by the ``prefix_provider``.

    Args:
        prefix_provider: A :obj:`str` prefix, Sequence[:obj:`str`] of prefixes, or sync or async callable that returns
            a prefix or sequence of prefixes. If ``None``, only the bot's mentions will be used.

    Example:

        .. code-block:: python

            # The below are all valid
            app = lightbulb.BotApp(prefix=lightbulb.when_mentioned_or("!"), ...)

            app = lightbulb.BotApp(prefix=lightbulb.when_mentioned_or(["!", "?"]), ...)

            # Using only mentions as the prefix
            app = lightbulb.BotApp(prefix=lightbulb.when_mentioned_or(None), ...)

            # Using with a get_prefix function
            def get_prefix(app, message):
                # Do something to get the prefixes
                return prefixes

            app = lightbulb.BotApp(prefix=lightbulb.when_mentioned_or(get_prefix), ...)
    """

    async def get_prefixes(app: BotApp, message: hikari.Message) -> t.Sequence[str]:
        me = app.get_me()
        assert me is not None
        mentions = [f"<@{me.id}> ", f"<@!{me.id}> "]

        if callable(prefix_provider):
            prefixes = prefix_provider(app, message)
            if inspect.iscoroutine(prefixes):
                prefixes = await prefixes
        else:
            prefixes = prefix_provider

        if isinstance(prefixes, str):
            return mentions + [prefixes]
        elif isinstance(prefixes, t.Sequence):
            return mentions + list(prefixes)
        return mentions

    return get_prefixes


# str is by definition a sequence of str so these type hints are correct
def _default_get_prefix(_: BotApp, __: hikari.Message, *, prefixes: t.Sequence[str]) -> t.Sequence[str]:
    return prefixes


class BotApp(hikari.GatewayBot):
    """
    A subclassed implementation of the :obj:`~hikari.impl.gateway_bot.GatewayBot` class containing a command
    handler. This should be instantiated instead of the superclass if you wish to use the command
    handler implementation provided.

    Args:
        token (:obj:`str`): The bot account's token.
        prefix (Optional[PrefixT]): The command prefix to use for prefix commands, or ``None`` if prefix commands
            will not be used.
        ignore_bots (:obj:`bool`): Whether or not prefix commands should ignore bots for invocation. Defaults
            to ``True``.
        owner_ids (Sequence[:obj:`int`]): The IDs of the users that own the bot. If not provided then it will be fetched
            by :obj:`~BotApp.fetch_owner_ids`.
        default_enabled_guilds (Union[:obj:`int`, Sequence[:obj:`int`]]): The guild(s) to create application commands
            in by default if no guilds are specified per-command. Defaults to an empty tuple.
        help_class (Optional[Type[:obj:`~.help_command.BaseHelpCommand`]]): Class to use for the bot's help command.
            Defaults to :obj:`~.help_command.DefaultHelpCommand`. If ``None``, no help command will be added
            to the bot by default.
        help_slash_command (:obj:`bool`): Whether or not the help command should be implemented as a slash command
            as well as a prefix command. Defaults to ``False``.
        delete_unbound_commands (:obj:`bool`): Whether or not the bot should delete application commands that it cannot
            find an implementation for when the bot starts. Defaults to ``True``.
        case_insensitive_prefix_commands (:obj:`bool`): Whether or not prefix command names should be case-insensitive.
            Defaults to ``False``.
        **kwargs (Any): Additional keyword arguments passed to the constructor of the :obj:`~hikari.impl.gateway_bot.GatewayBot`
            class.
    """

    __slots__ = (
        "get_prefix",
        "ignore_bots",
        "owner_ids",
        "application",
        "d",
        "_prefix_commands",
        "_slash_commands",
        "_message_commands",
        "_user_commands",
        "_plugins",
        "_checks",
        "extensions",
        "_current_extension",
        "default_enabled_guilds",
        "_help_command",
        "_delete_unbound_commands",
        "_case_insensitive_prefix_commands",
        "_running_tasks",
    )

    def __init__(
        self,
        token: str,
        prefix: t.Optional[PrefixT] = None,
        ignore_bots: bool = True,
        owner_ids: t.Sequence[int] = (),
        default_enabled_guilds: t.Union[int, t.Sequence[int]] = (),
        help_class: t.Optional[t.Type[help_command_.BaseHelpCommand]] = help_command_.DefaultHelpCommand,
        help_slash_command: bool = False,
        delete_unbound_commands: bool = True,
        case_insensitive_prefix_commands: bool = False,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(token, **kwargs)
        # The prefix command handler expects an iterable to be returned from the get_prefix function,
        # so we have to wrap a single string prefix in a list here.
        if prefix is not None:
            prefix = [prefix] if isinstance(prefix, str) else prefix
            if isinstance(prefix, t.Sequence):
                # Create the default get prefix from the passed-in prefixes if a get_prefix function
                # was not provided
                prefix = functools.partial(_default_get_prefix, prefixes=prefix)
            self.get_prefix: t.Callable[
                [BotApp, hikari.Message], t.Union[t.Sequence[str], t.Coroutine[t.Any, t.Any, t.Sequence[str]]]
            ] = prefix

        self._delete_unbound_commands = delete_unbound_commands
        self._case_insensitive_prefix_commands = case_insensitive_prefix_commands

        self.ignore_bots: bool = ignore_bots
        """Whether or not other bots will be ignored when invoking prefix commands."""
        self.owner_ids: t.Sequence[int] = owner_ids
        """The owner ID(s) for the owner(s) of the bot account."""
        self.default_enabled_guilds: t.Sequence[int] = (
            (default_enabled_guilds,) if isinstance(default_enabled_guilds, int) else default_enabled_guilds
        )
        """The default guilds that application commands will be enabled in."""

        self.application: t.Optional[hikari.Application] = None
        """The :obj:`~hikari.applications.Application` for the bot account. This will always be ``None`` before the bot has logged in."""

        self.d: data_store.DataStore = data_store.DataStore()
        """A :obj:`~.utils.data_store.DataStore` instance enabling storage of custom data without subclassing."""

        self.extensions: t.List[str] = []
        """A list of the currently loaded extensions."""
        self._current_extension: t.Optional[_ExtensionT] = None

        self._prefix_commands: t.MutableMapping[str, commands.prefix.PrefixCommand] = (
            {} if not case_insensitive_prefix_commands else CIMultiDict()  # type: ignore
        )
        self._slash_commands: t.MutableMapping[str, commands.slash.SlashCommand] = {}
        self._message_commands: t.MutableMapping[str, commands.message.MessageCommand] = {}
        self._user_commands: t.MutableMapping[str, commands.user.UserCommand] = {}

        self._plugins: t.MutableMapping[str, plugins_.Plugin] = {}

        self._checks: t.List[t.Union[checks.Check, checks._ExclusiveCheck]] = []

        self._help_command: t.Optional[help_command_.BaseHelpCommand] = None
        if help_class is not None:
            help_cmd_types: t.List[t.Type[commands.base.Command]] = []

            if prefix is not None:
                help_cmd_types.append(commands.prefix.PrefixCommand)

            if help_slash_command:
                help_cmd_types.append(commands.slash.SlashCommand)

            if help_cmd_types:
                self._help_command = help_class(self)

                @decorators.option(
                    "obj", "Object to get help for", required=False, modifier=commands.base.OptionModifier.CONSUME_REST
                )
                @decorators.command("help", "Get help information for the bot", auto_defer=True)
                @decorators.implements(*help_cmd_types)
                async def __default_help(ctx: context_.base.Context) -> None:
                    assert self._help_command is not None
                    await self._help_command.send_help(ctx, ctx.options.obj)

                self.command(__default_help)

        # We need to store created tasks internally to ensure that they do not
        # get destroyed mid-execution. See asyncio.create_task documentation for more.
        self._running_tasks: t.List[asyncio.Task[t.Any]] = []

        if prefix is not None:
            self.subscribe(hikari.MessageCreateEvent, self.handle_message_create_for_prefix_commands)
        self.subscribe(hikari.StartedEvent, self._manage_application_commands)
        self.subscribe(hikari.InteractionCreateEvent, self.handle_interaction_create_for_application_commands)
        self.subscribe(hikari.InteractionCreateEvent, self.handle_interaction_create_for_autocomplete)

    def create_task(self, coro: t.Awaitable[t.Any], *, name: t.Optional[str] = None) -> asyncio.Task[t.Any]:
        """
        Wrap the given awaitable into an :obj:`asyncio.Task` and schedule its execution. This
        method functions the same as :meth:`asyncio.create_task`, but keeps a reference to the task
        alive until execution is completed to ensure the task is not destroyed mid-execution.

        Args:
            coro (Awaitable[Any]): Coroutine to wrap into a task.

        Keyword Args:
            name (Optional[:obj:`str`]): The name of the task. Not required, defaults to ``None``.

        Returns:
            :obj:`asyncio.Task`: Created task object.

        .. versionadded:: 2.2.0
        """
        task: asyncio.Task[None] = asyncio.create_task(coro, name=name)  # type: ignore[arg-type]
        self._running_tasks.append(task)
        task.add_done_callback(lambda task_: self._running_tasks.remove(task_))
        return task

    @property
    def help_command(self) -> t.Optional[help_command_.BaseHelpCommand]:
        """The current help command instance registered to the bot."""
        return self._help_command

    @help_command.setter
    def help_command(self, val: help_command_.BaseHelpCommand) -> None:
        self._help_command = val

    @property
    def prefix_commands(self) -> t.MutableMapping[str, commands.prefix.PrefixCommand]:
        """Mapping of command name to command object containing all prefix commands registered to the bot."""
        return self._prefix_commands

    @property
    def slash_commands(self) -> t.MutableMapping[str, commands.slash.SlashCommand]:
        """Mapping of command name to command object containing all slash commands registered to the bot."""
        return self._slash_commands

    @property
    def message_commands(self) -> t.MutableMapping[str, commands.message.MessageCommand]:
        """Mapping of command name to command object containing all message commands registered to the bot."""
        return self._message_commands

    @property
    def user_commands(self) -> t.MutableMapping[str, commands.user.UserCommand]:
        """Mapping of command name to command object containing all user commands registered to the bot."""
        return self._user_commands

    @property
    def plugins(self) -> t.MutableMapping[str, plugins_.Plugin]:
        """Mapping of plugin name to plugin object containing all plugins registered to the bot."""
        return self._plugins

    def _add_command_to_correct_attr(self, command: commands.base.Command) -> None:
        if isinstance(command, commands.prefix.PrefixCommand):
            for item in [command.name, *command.aliases]:
                if item in self._prefix_commands:
                    raise errors.CommandAlreadyExists(
                        f"A prefix command with name or alias {item!r} is already registered."
                    )
            for item in [command.name, *command.aliases]:
                self._prefix_commands[item] = command
        elif isinstance(command, commands.slash.SlashCommand):
            if command.name in self._slash_commands:
                raise errors.CommandAlreadyExists(f"A slash command with name {command.name!r} is already registered.")
            self._slash_commands[command.name] = command
        elif isinstance(command, commands.message.MessageCommand):
            if command.name in self._message_commands:
                raise errors.CommandAlreadyExists(
                    f"A message command with name {command.name!r} is already registered."
                )
            self._message_commands[command.name] = command
        elif isinstance(command, commands.user.UserCommand):
            if command.name in self._user_commands:
                raise errors.CommandAlreadyExists(f"A user command with name {command.name!r} is already registered.")
            self._user_commands[command.name] = command

    def _get_application_command(
        self, interaction: hikari.CommandInteraction
    ) -> t.Optional[commands.base.ApplicationCommand]:
        if interaction.command_type is hikari.CommandType.SLASH:
            return self.get_slash_command(interaction.command_name)
        elif interaction.command_type is hikari.CommandType.USER:
            return self.get_user_command(interaction.command_name)
        elif interaction.command_type is hikari.CommandType.MESSAGE:
            return self.get_message_command(interaction.command_name)
        return None

    async def _manage_application_commands(self, _: hikari.StartingEvent) -> None:
        if self.application is None:
            self.application = await self.rest.fetch_application()

        try:
            await internal.manage_application_commands(self)
        except hikari.ForbiddenError as exc:
            error_msg = str(exc)
            match = _APPLICATION_CMD_ERROR_REGEX.search(error_msg)
            guild_id = "unknown" if match is None else match.group(1)
            raise errors.ApplicationCommandCreationFailed(
                f"Application command creation failed for guild {guild_id!r}. Is your bot in the guild and was it invited with the 'applications.commands' scope?"
            ) from exc
        finally:
            await self.dispatch(events.LightbulbStartedEvent(app=self))

    async def sync_application_commands(self) -> None:
        """
        Sync all application commands registered to the bot with discord.

        Returns:
            ``None``

        .. versionadded:: 2.2.0
        """
        await internal.manage_application_commands(self)

    @staticmethod
    def _get_events_for_application_command(
        command: commands.base.ApplicationCommand,
    ) -> t.Tuple[
        t.Type[events.CommandInvocationEvent], t.Type[events.CommandCompletionEvent], t.Type[events.CommandErrorEvent]
    ]:
        for k in APPLICATION_COMMANDS_EVENTS_MAPPING:
            if isinstance(command, k):
                return APPLICATION_COMMANDS_EVENTS_MAPPING[k]
        raise TypeError("Application command type not recognised")

    def print_banner(  # type: ignore[override]
        self,
        banner: t.Optional[str],
        allow_color: bool,
        force_color: bool,
        extra_args: t.Optional[t.Dict[str, str]] = None,
    ) -> None:
        super().print_banner(banner, allow_color, force_color)
        if banner == "hikari":
            sys.stdout.write("Thank you for using lightbulb!\n")

    def load_extensions(self, *extensions: str) -> None:
        """
        Load external extension(s) into the bot. Extension name follows the format ``<directory>.<filename>``
        Each extension **must** contain a function ``load`` which takes a single argument which will be the
        ``BotApp`` instance you are loading the extension into.

        Args:
            extensions (:obj:`str`): The name of the extension(s) to load.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.ExtensionAlreadyLoaded`: If the extension has already been loaded.
            :obj:`~.errors.ExtensionMissingLoad`: If the extension to be loaded does not contain a ``load`` function.
            :obj:`~.errors.ExtensionNotFound`: If the extension to be loaded does not exist.
        """
        if len(extensions) > 1 or not extensions:
            for extension in extensions:
                self.load_extensions(extension)
            return
        extension = extensions[0]

        if extension in self.extensions:
            raise errors.ExtensionAlreadyLoaded(f"Extension {extension!r} is already loaded.")

        spec = util.find_spec(extension)
        if spec is None:
            raise errors.ExtensionNotFound(f"No extension by the name {extension!r} was found")

        module = importlib.import_module(extension)

        ext = t.cast(_ExtensionT, module)
        self._current_extension = ext

        if not hasattr(module, "load"):
            raise errors.ExtensionMissingLoad(f"Extension {extension!r} is missing a load function")
        else:
            ext.load(self)
            self.extensions.append(extension)
            _LOGGER.info("Extension loaded %r", extension)
        self._current_extension = None

    def unload_extensions(self, *extensions: str) -> None:
        """
        Unload external extension(s) from the bot. This method relies on a function, ``unload``
        existing in the extensions which the bot will use to remove all commands and/or plugins
        from the bot.

        Args:
            extensions (:obj:`str`): The name of the extension(s) to unload.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.ExtensionNotLoaded`: If the extension has not been loaded.
            :obj:`~.errors.ExtensionMissingUnload`: If the extension does not contain an ``unload`` function.
            :obj:`~.errors.ExtensionNotFound`: If the extension to be unloaded does not exist.
        """
        if len(extensions) > 1 or not extensions:
            for extension in extensions:
                self.unload_extensions(extension)
            return
        extension = extensions[0]

        if extension not in self.extensions:
            raise errors.ExtensionNotLoaded(f"Extension {extension!r} is not loaded.")

        try:
            module = importlib.import_module(extension)
        except ModuleNotFoundError:
            raise errors.ExtensionNotFound(f"No extension by the name {extension!r} was found") from None

        ext = t.cast(_ExtensionT, module)
        self._current_extension = ext

        if not hasattr(module, "unload"):
            raise errors.ExtensionMissingUnload(f"Extension {extension!r} is missing an unload function")
        else:
            ext.unload(self)
            self.extensions.remove(extension)
            del sys.modules[extension]
            _LOGGER.info("Extension unloaded %r", extension)
        self._current_extension = None

    def reload_extensions(self, *extensions: str) -> None:
        """
        Reload bot extension(s). This method is atomic and so the bot will
        revert to the previous loaded state if an extension encounters a problem
        during unloading or loading.

        Args:
            extensions (:obj:`str`): The name of the extension(s) to be reloaded.

        Returns:
            ``None``
        """
        if len(extensions) > 1 or not extensions:
            for extension in extensions:
                self.reload_extensions(extension)
            return
        extension = extensions[0]
        try:
            old = sys.modules[extension]
        except KeyError:
            raise errors.ExtensionNotLoaded(f"Extension {extension!r} is not loaded.")
        try:
            self.unload_extensions(extension)
            self.load_extensions(extension)
        except Exception as e:
            sys.modules[extension] = old
            if not isinstance(e, errors.ExtensionAlreadyLoaded):
                self.load_extensions(extension)
            raise e
        else:
            del old

    def load_extensions_from(
        self, *paths: t.Union[str, pathlib.Path], recursive: bool = False, must_exist: bool = True
    ) -> None:
        """
        Load all external extensions from the given directories. Files that begin with an underscore ( _ ) are ignored.
        Every extension **must** contain a function ``load`` which takes a single argument which will be the ``BotApp``
        instance you are loading the extension into.

        Args:
            *paths (Union[:obj:`str`, :obj:`pathlib.Path`]): The directories to load extensions from. These can be
                relative or absolute directories. In the case that a directory is absolute, the extension name will
                be set as though the directory was relative.

        Keyword Args:
            recursive (:obj:`bool`): Whether to search the directories recursively. Defaults to False.
            must_exist (:obj:`bool`): Whether all directories must exist before extensions can be loaded. If this is
                False and a directory does not exist, the directory will be ignored. If this is True, a
                :obj:`FileNotFoundError` is thrown if any directory does not exist. Defaults to True.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.ExtensionAlreadyLoaded`: If the extension has already been loaded.
            :obj:`~.errors.ExtensionMissingLoad`: If the extension to be loaded does not contain a ``load`` function.
            :obj:`~.errors.ExtensionNotFound`: If the extension to be loaded does not exist.
            :obj:`FileNotFoundError`: If any directory to load extensions from does not exist and ``must_exist``
                is True.
            :obj:`ValueError`: If the path provided is not relative to the current working directory.
        """
        if len(paths) > 1 or not paths:
            for path_ in paths:
                self.load_extensions_from(path_, recursive=recursive, must_exist=must_exist)
            return

        path = paths[0]

        if isinstance(path, str):
            path = pathlib.Path(path)

        try:
            path = path.resolve().relative_to(pathlib.Path.cwd())
        except ValueError:
            raise ValueError(f"'{path}' must be relative to the working directory") from None

        if not path.is_dir():
            if must_exist:
                raise FileNotFoundError(f"'{path}' is not an existing directory")
            return

        glob = path.rglob if recursive else path.glob
        for ext_path in glob("[!_]*.py"):
            ext = str(ext_path.with_suffix("")).replace(os.sep, ".")
            self.load_extensions(ext)

    async def fetch_owner_ids(self) -> t.Sequence[hikari.Snowflakeish]:
        """
        Fetch the bot's owner IDs, or return the given owner IDs on instantiation if provided.

        Returns:
            Sequence[Snowflakeish]: The IDs of the bot's owners.
        """
        if self.owner_ids:
            return self.owner_ids

        self.application = self.application or await self.rest.fetch_application()

        owner_ids = []
        if self.application.owner is not None:
            owner_ids.append(self.application.owner.id)
        if self.application.team is not None:
            owner_ids.extend([member_id for member_id in self.application.team.members])
        return owner_ids

    async def maybe_dispatch_error_event(
        self,
        event: events.CommandErrorEvent,
        priority_handlers: t.Sequence[
            t.Union[None, t.Callable[[events.CommandErrorEvent], t.Coroutine[t.Any, t.Any, t.Optional[bool]]]]
        ],
    ) -> bool:
        """
        Attempts to handle the event first using the given ``priority_handlers``, falling back to dispatching
        the given event to global listeners.

        Args:
            event (:obj:`~.events.CommandErrorEvent`): Event to attempt to handle.
            priority_handlers (Sequence[Union[``None``, ListenerT)]]: Handlers to attempt to use to handle the
                event before falling back to global event listeners.

        Returns:
            :obj:`bool`: Whether or not the given event was handled or dispatched successfully.
        """
        handled = False
        for listener in priority_handlers:
            if handled:
                break
            if listener is not None:
                handled = bool(await listener(event))

        if not handled:
            if self.get_listeners(type(event), polymorphic=True):
                await self.dispatch(event)
                handled = True
            elif self.get_listeners(events.CommandErrorEvent, polymorphic=True):
                await self.dispatch(event)
                handled = True

        return handled

    @t.overload
    def check(self, check: t.Union[checks.Check, CheckCoroT]) -> checks.Check:
        ...

    @t.overload
    def check(self) -> t.Callable[[CheckCoroT], checks.Check]:
        ...

    def check(
        self,
        check: t.Optional[t.Union[checks.Check, CheckCoroT, checks._ExclusiveCheck]] = None,
    ) -> t.Union[checks.Check, t.Callable[[CheckCoroT], checks.Check]]:
        """
        Adds a :obj:`~.checks.Check` object or check function the bot's checks. This method can be used as a
        first or second order decorator, or called manually with the :obj:`~.checks.Check` instance or function to
        add as a check. If a function is decorated or passed in then it will be wrapped in a :obj:`~.checks.Check`
        object before it is added to the bot.
        """
        if check is not None:
            if not isinstance(check, checks.Check):
                check = checks.Check(check)
            self._checks.append(check)

            check.add_to_object_hook(self)

            return check

        def decorate(
            check_func: t.Callable[[context_.base.Context], t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]]
        ) -> checks.Check:
            new_check = checks.Check(check_func)
            self._checks.append(new_check)

            new_check.add_to_object_hook(self)

            return new_check

        return decorate

    def get_prefix_command(self, name: str) -> t.Optional[commands.prefix.PrefixCommand]:
        """
        Gets the prefix command with the given name, or ``None`` if no command with that name was found.

        Args:
            name (:obj:`str`): Name of the prefix command to get.

        Returns:
            Optional[:obj:`~.commands.prefix.PrefixCommand`]: Prefix command object with the given name, or ``None``
                if not found.
        """
        parts = name.split()
        if len(parts) == 1:
            return self._prefix_commands.get(name)

        maybe_group = self._prefix_commands.get(parts.pop(0))
        if not isinstance(maybe_group, commands.prefix.PrefixCommandGroup):
            return None

        this: t.Optional[
            t.Union[
                commands.prefix.PrefixCommandGroup, commands.prefix.PrefixSubGroup, commands.prefix.PrefixSubCommand
            ]
        ] = maybe_group
        for part in parts:
            if this is None or isinstance(this, commands.prefix.PrefixSubCommand):
                return None
            this = this.get_subcommand(part)

        return this

    def get_slash_command(self, name: str) -> t.Optional[commands.slash.SlashCommand]:
        """
        Gets the slash command with the given name, or ``None`` if no command with that name was found.

        Args:
            name (:obj:`str`): Name of the slash command to get.

        Returns:
            Optional[:obj:`~.commands.slash.SlashCommand`]: Slash command object with the given name, or ``None``
                if not found.
        """
        return self._slash_commands.get(name)

    def get_message_command(self, name: str) -> t.Optional[commands.message.MessageCommand]:
        """
        Gets the message command with the given name, or ``None`` if no command with that name was found.

        Args:
            name (:obj:`str`): Name of the message command to get.

        Returns:
            Optional[:obj:`~.commands.message.MessageCommand`]: Message command object with the given name, or ``None``
                if not found.
        """
        return self._message_commands.get(name)

    def get_user_command(self, name: str) -> t.Optional[commands.user.UserCommand]:
        """
        Gets the user command with the given name, or ``None`` if no command with that name was found.

        Args:
            name (:obj:`str`): Name of the user command to get.

        Returns:
            Optional[:obj:`~.commands.user.UserCommand`]: User command object with the given name, or ``None``
                if not found.
        """
        return self._user_commands.get(name)

    @t.overload
    def command(self, cmd_like: commands.base.CommandLike) -> commands.base.CommandLike:
        ...

    @t.overload
    def command(self) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
        ...

    def command(
        self, cmd_like: t.Optional[commands.base.CommandLike] = None
    ) -> t.Union[commands.base.CommandLike, t.Callable[[commands.base.CommandLike], commands.base.CommandLike]]:
        """
        Adds a :obj:`~.commands.base.CommandLike` object as a command to the bot. This method can be used as a
        first or second order decorator, or called manually with the :obj:`~.commands.CommandLike` instance to
        add as a command.
        """
        if cmd_like is not None:
            commands_to_impl: t.Sequence[t.Type[commands.base.Command]] = getattr(
                cmd_like.callback, "__cmd_types__", []
            )
            _LOGGER.debug(
                "Registering command %r. Requested types are: %s",
                cmd_like.name,
                ",".join(c.__name__ for c in commands_to_impl),
            )
            for command_cls in commands_to_impl:
                cmd = command_cls(self, cmd_like)

                if cmd.is_subcommand:
                    continue

                cmd._validate_attributes()

                self._add_command_to_correct_attr(cmd)
            return cmd_like

        def decorate(cmd_like_: commands.base.CommandLike) -> commands.base.CommandLike:
            self.command(cmd_like_)
            return cmd_like_

        return decorate

    def remove_command(self, command: t.Union[commands.base.Command, commands.base.CommandLike]) -> None:
        """
        Removes a command or command-like object from the bot.

        Args:
            command (Union[:obj:`~.commands.base.Command`, :obj:`~.commands.base.CommandLike`): Command or
                command-like object to remove from the bot.

        Returns:
            ``None``
        """
        _LOGGER.debug("Removing command %r (%s)", command.name, command.__class__.__name__)
        if isinstance(command, commands.base.CommandLike):
            self._remove_commandlike(command)
            return

        if isinstance(command, commands.prefix.PrefixCommand):
            for item in [command.name, *command.aliases]:
                self._prefix_commands.pop(item, None)
        elif isinstance(command, commands.slash.SlashCommand):
            self._slash_commands.pop(command.name, None)
        elif isinstance(command, commands.message.MessageCommand):
            self._message_commands.pop(command.name, None)
        elif isinstance(command, commands.user.UserCommand):
            self._user_commands.pop(command.name, None)

    def _remove_commandlike(self, cmd_like: commands.base.CommandLike) -> None:
        commands_to_remove: t.List[t.Optional[commands.base.Command]] = []
        cmd_types: t.Sequence[t.Type[commands.base.Command]] = getattr(cmd_like.callback, "__cmd_types__", [])
        for cmd_type in cmd_types:
            if issubclass(cmd_type, commands.prefix.PrefixCommand):
                commands_to_remove.append(self.get_prefix_command(cmd_like.name))
            elif issubclass(cmd_type, commands.slash.SlashCommand):
                commands_to_remove.append(self.get_slash_command(cmd_like.name))
            elif issubclass(cmd_type, commands.message.MessageCommand):
                commands_to_remove.append(self.get_message_command(cmd_like.name))
            elif issubclass(cmd_type, commands.user.UserCommand):
                commands_to_remove.append(self.get_user_command(cmd_like.name))

        for command in filter(None, commands_to_remove):
            self.remove_command(command)

    def get_plugin(self, name: str) -> t.Optional[plugins_.Plugin]:
        """
        Gets the plugin with the given name, or ``None`` if no plugin with that name was found.

        Args:
            name (:obj:`str`): Name of the plugin to get.

        Returns:
            Optional[:obj:`~.plugins.Plugin`]: Plugin object with the given name, or ``None`` if not found.
        """
        return self._plugins.get(name)

    def add_plugin(self, plugin: plugins_.Plugin) -> None:
        """
        Registers a plugin to the bot, adding all commands and listeners present
        in the plugin.

        Args:
            plugin (:obj:`~.plugins.Plugin`): Plugin to register to the bot.

        Returns:
            ``None``
        """
        plugin.app = self
        for command in plugin._all_commands:
            self._add_command_to_correct_attr(command)
        for event, listeners in plugin._listeners.items():
            for listener in listeners:
                self.subscribe(event, listener)
        _LOGGER.debug("Plugin registered %r", plugin.name)
        self._plugins[plugin.name] = plugin

    def remove_plugin(self, plugin_or_name: t.Union[plugins_.Plugin, str]) -> None:
        """
        Unregisters a plugin from the bot, removing all commands and listeners
        present in the plugin.

        Args:
            plugin_or_name (Union[:obj:`~.plugins.Plugin`, :obj:`str`]): Plugin or name of the plugin
                to unregister from the bot.

        Returns:
            ``None``
        """
        plugin: t.Optional[t.Union[plugins_.Plugin, str]] = plugin_or_name
        if isinstance(plugin, str):
            plugin = self.get_plugin(plugin)

        if plugin is None:
            return

        assert isinstance(plugin, plugins_.Plugin)

        for command in plugin._raw_commands:
            self.remove_command(command)
        for event, listeners in plugin._listeners.items():
            for listener in listeners:
                self.unsubscribe(event, listener)

        if plugin._remove_hook is not None:
            maybe_coro = plugin._remove_hook()
            if inspect.iscoroutine(maybe_coro):
                self.create_task(maybe_coro)

        self._plugins.pop(plugin.name, None)
        _LOGGER.debug("Plugin removed %r", plugin.name)

    async def purge_application_commands(self, *guild_ids: hikari.Snowflakeish, global_commands: bool = False) -> None:
        """
        Purges all application commands from the guilds with the specified IDs, and all the global application
        commands if ``global_commands`` is ``True``. Useful if you want to teardown old slash commands from
        the bot and cannot be bothered to fetch them individually yourself. If neither `guild_ids` nor `global_commands`
        is specified then this method will do nothing.

        Args:
            *guild_ids (:obj:`hikari.Snowflakeish`): IDs for the guilds to purge application commands from.

        Keyword Args:
            global_commands (:obj:`bool`): Whether or not to purge global slash commands from the bot.

        Returns:
            ``None``
        """
        assert self.application is not None
        if global_commands:
            await self.rest.set_application_commands(self.application, ())
        if guild_ids:
            for guild_id in guild_ids:
                await self.rest.set_application_commands(self.application, (), guild_id)

    async def get_prefix_context(
        self,
        event: hikari.MessageCreateEvent,
        cls: t.Type[context_.prefix.PrefixContext] = context_.prefix.PrefixContext,
    ) -> t.Optional[context_.prefix.PrefixContext]:
        """
        Get the :obj:`~.context.prefix.PrefixContext` instance for the given event, or ``None`` if
        no context could be created.

        Args:
            event (:obj:`~hikari.events.message_events.MessageCreateEvent`): Event to get the prefix context for.
            cls (Type[:obj:`~.context.prefix.PrefixContext`]): Context class to instantiate. Defaults to
                :obj:`~.context.prefix.PrefixContext`.

        Returns:
            Optional[:obj:`~.context.prefix.PrefixContext`]: Prefix context instance for the given event.
        """
        assert event.message.content is not None

        prefixes = self.get_prefix(self, event.message)
        if inspect.iscoroutine(prefixes):
            prefixes = await prefixes
        prefixes = t.cast(t.Sequence[str], prefixes)

        if isinstance(prefixes, str):
            prefixes = [prefixes]
        prefixes = sorted(prefixes, key=len, reverse=True)

        invoked_prefix = None
        for prefix in prefixes:
            if event.message.content.startswith(prefix):
                invoked_prefix = prefix
                break

        if invoked_prefix is None:
            return None

        new_content = event.message.content[len(invoked_prefix) :]
        if not new_content or new_content.isspace():
            return None

        split_content = new_content.split(maxsplit=1)
        invoked_with, args = split_content[0], "".join(split_content[1:])

        if not invoked_with:
            return None

        command = self.get_prefix_command(invoked_with)
        ctx = cls(self, event, command, invoked_with, invoked_prefix)
        if ctx.command is not None:
            ctx._parser = (ctx.command.parser or parser.Parser)(ctx, args)
        return ctx

    async def process_prefix_commands(self, context: context_.prefix.PrefixContext) -> None:
        """
        Invokes the appropriate command for the given context.

        Args:
            context (:obj:`.context.prefix.PrefixContext`): Context to invoke the command under.

        Returns:
            ``None``
        """
        if context.command is None:
            raise errors.CommandNotFound(
                f"A command with name or alias {context.invoked_with!r} does not exist",
                invoked_with=context.invoked_with,
            )

        await context.invoke()

    async def handle_message_create_for_prefix_commands(self, event: hikari.MessageCreateEvent) -> None:
        """
        Prefix command :obj:`~hikari.events.message_events.MessageCreateEvent` listener. This handles fetching the
        context, dispatching events, and invoking the appropriate command.

        Args:
            event (:obj:`~hikari.events.message_events.MessageCreateEvent`): Event that prefix commands will be
                processed for.

        Returns:
            ``None``
        """
        if self.ignore_bots and not event.is_human:
            return

        if not event.message.content:
            return

        context = await self.get_prefix_context(event)
        if context is None:
            return

        if context.command is not None:
            await self.dispatch(events.PrefixCommandInvocationEvent(app=self, command=context.command, context=context))

        try:
            await self.process_prefix_commands(context)
        except Exception as exc:
            new_exc = exc
            if not isinstance(exc, errors.LightbulbError):
                assert context.command is not None
                new_exc = errors.CommandInvocationError(
                    f"An error occurred during command {context.command.name!r} invocation", original=exc
                )
            assert isinstance(new_exc, errors.LightbulbError)
            error_event = events.PrefixCommandErrorEvent(app=self, exception=new_exc, context=context)
            handled = await self.maybe_dispatch_error_event(
                error_event,
                [
                    getattr(context.command, "error_handler", None),
                    getattr(context.command.plugin, "_error_handler", None) if context.command is not None else None,
                ],
            )

            if not handled:
                raise new_exc
        else:
            assert context.command is not None
            await self.dispatch(events.PrefixCommandCompletionEvent(app=self, command=context.command, context=context))

    async def get_slash_context(
        self,
        event: hikari.InteractionCreateEvent,
        command: commands.slash.SlashCommand,
        cls: t.Type[context_.slash.SlashContext] = context_.slash.SlashContext,
    ) -> context_.slash.SlashContext:
        """
        Get the :obj:`~.context.slash.SlashContext` instance for the given event.

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event to get the slash context for.
            command (:obj:`~.commands.slash.SlashCommand`): Command that the context is for.
            cls (Type[:obj:`~.context.slash.SlashContext`]): Context class to instantiate. Defaults to
                :obj:`~.context.slash.SlashContext`.

        Returns:
            :obj:`~.context.slash.SlashContext`: Slash context instance for the given event.
        """
        return cls(self, event, command)

    async def get_message_context(
        self,
        event: hikari.InteractionCreateEvent,
        command: commands.message.MessageCommand,
        cls: t.Type[context_.message.MessageContext] = context_.message.MessageContext,
    ) -> context_.message.MessageContext:
        """
        Get the :obj:`~.context.message.MessageContext` instance for the given event.

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event to get the message context
                for.
            command (:obj:`~.commands.message.MessageCommand`): Command that the context is for.
            cls (Type[:obj:`~.context.message.MessageContext`]): Context class to instantiate. Defaults to
                :obj:`~.context.message.MessageContext`.

        Returns:
            :obj:`~.context.message.MessageContext`: Message context instance for the given event.
        """
        return cls(self, event, command)

    async def get_user_context(
        self,
        event: hikari.InteractionCreateEvent,
        command: commands.user.UserCommand,
        cls: t.Type[context_.user.UserContext] = context_.user.UserContext,
    ) -> context_.user.UserContext:
        """
        Get the :obj:`~.context.user.UserContext` instance for the given event.

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event to get the user context for.
            command (:obj:`~.commands.slash.SlashCommand`): Command that the context is for.
            cls (Type[:obj:`~.context.user.UserContext`]): Context class to instantiate. Defaults to
                :obj:`~.context.user.UserContext`.

        Returns:
            :obj:`~.context.user.UserContext`: User context instance for the given event.
        """
        return cls(self, event, command)

    async def get_application_command_context(
        self, event: hikari.InteractionCreateEvent
    ) -> t.Optional[context_.base.ApplicationContext]:
        """
        Get the appropriate subclass instance of :obj:`~.context.base.Application` for the given event.

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event to get the context for.

        Returns:
            :obj:`~.context.base.ApplicationContext`: Context instance for the given event.
        """
        assert isinstance(event.interaction, hikari.CommandInteraction)
        cmd = self._get_application_command(event.interaction)
        if cmd is None:
            return None

        if isinstance(cmd, commands.slash.SlashCommand):
            return await self.get_slash_context(event, cmd)
        elif isinstance(cmd, commands.user.UserCommand):
            return await self.get_user_context(event, cmd)
        elif isinstance(cmd, commands.message.MessageCommand):
            return await self.get_message_context(event, cmd)
        return None

    async def invoke_application_command(self, context: context_.base.ApplicationContext) -> None:
        """
        Invokes the appropriate application command for the given context and handles event
        dispatching for the command invocation.

        Args:
            context (:obj:`~.context.base.ApplicationContext`): Context to invoke application commands under.

        Returns:
            ``None``
        """
        cmd_events = self._get_events_for_application_command(context.command)
        await self.dispatch(cmd_events[0](app=self, command=context.command, context=context))

        try:
            await context.invoke()
        except Exception as exc:
            new_exc = exc
            if not isinstance(exc, errors.LightbulbError):
                new_exc = errors.CommandInvocationError(
                    f"An error occurred during command {context.command.name!r} invocation", original=exc
                )
            assert isinstance(new_exc, errors.LightbulbError)
            error_event = cmd_events[2](app=self, exception=new_exc, context=context)
            handled = await self.maybe_dispatch_error_event(
                error_event,
                [
                    getattr(context.command, "error_handler", None),
                    getattr(context.command.plugin, "_error_handler", None),
                ],
            )

            if not handled:
                raise new_exc
        else:
            await self.dispatch(cmd_events[1](app=self, command=context.command, context=context))

    async def handle_interaction_create_for_application_commands(self, event: hikari.InteractionCreateEvent) -> None:
        """
        Application command :obj:`~hikari.events.interaction_events.InteractionCreateEvent` listener. This handles
        fetching the context, dispatching events, and invoking the appropriate command.

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event that application commands
                will be processed for.

        Returns:
            ``None``
        """
        if not isinstance(event.interaction, hikari.CommandInteraction):
            return

        context = await self.get_application_command_context(event)
        if context is None:
            return

        await self.invoke_application_command(context)

    async def handle_interaction_create_for_autocomplete(self, event: hikari.InteractionCreateEvent) -> None:
        """
        Autocomplete :obj:`~hikari.events.interaction_events.InteractionCreateEvent` listener. This handles resolving
        the function to use for autocompletion, response conversion into :obj:`~hikari.commands.CommandChoice` and
        responding to the interaction with the provided options

        Args:
            event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): Event that autocomplete
                will be processed for.

        Returns:
            ``None``
        """
        if not isinstance(event.interaction, hikari.AutocompleteInteraction):
            return

        assert event.interaction.command_type is hikari.CommandType.SLASH
        assert event.interaction.options is not None and len(event.interaction.options) > 0

        get_focused: t.Callable[
            [t.Sequence[hikari.AutocompleteInteractionOption]], hikari.AutocompleteInteractionOption
        ] = lambda opts: next(filter(lambda o: o.is_focused, opts), opts[0])

        def flatten_command_option(
            opt: hikari.AutocompleteInteractionOption,
        ) -> t.Tuple[hikari.AutocompleteInteractionOption, t.Sequence[str]]:
            current = opt
            name = [current.name]

            while current.type in (hikari.OptionType.SUB_COMMAND, hikari.OptionType.SUB_COMMAND_GROUP):
                assert current.options is not None
                current = get_focused(current.options)
                name.append(current.name)

            return current, name

        option, full_name = flatten_command_option(get_focused(event.interaction.options))

        cmd = self.get_slash_command(event.interaction.command_name)
        if cmd is None:
            return

        for part in full_name[:-1]:
            if not isinstance(cmd, commands.slash.SlashGroupMixin):
                return

            cmd = cmd.get_subcommand(part)
            if cmd is None:
                return

        callback = cmd._initialiser._autocomplete_callbacks.get(full_name[-1])
        if callback is None:
            return

        # Invoke the autocomplete callback
        response = await callback(option, event.interaction)
        if not response:
            await event.interaction.create_response([])
            return

        def convert_response_value(
            val: t.Union[str, int, float, hikari.api.AutocompleteChoiceBuilder]
        ) -> hikari.api.AutocompleteChoiceBuilder:
            if isinstance(val, (str, int, float)):
                return hikari.impl.AutocompleteChoiceBuilder(name=str(val), value=val)

            return val

        resp_to_send: t.List[hikari.api.AutocompleteChoiceBuilder] = []
        if isinstance(response, (str, int, float, hikari.api.AutocompleteChoiceBuilder)):
            resp_to_send.append(convert_response_value(response))
        elif isinstance(response, collections.abc.Sequence):
            for item in response:
                resp_to_send.append(convert_response_value(item))
        else:
            _LOGGER.error("Invalid response returned from autocomplete handler %r", callback.__name__)  # type: ignore[unreachable]

        try:
            await event.interaction.create_response(resp_to_send)
        except hikari.NotFoundError as e:
            _LOGGER.debug("Failed sending autocomplete response", exc_info=(type(e), e, e.__traceback__))



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "ApplicationCommand",
    "ApplicationCommandCreationFailed",
    "ApplicationContext",
    "BangBangCooldownAlgorithm",
    "BaseConverter",
    "BaseHelpCommand",
    "BaseParser",
    "BooleanConverter",
    "BotApp",
    "BotMissingRequiredPermission",
    "BotOnly",
    "Bucket",
    "ChannelBucket",
    "Check",
    "CheckFailure",
    "ColorConverter",
    "ColourConverter",
    "Command",
    "CommandAlreadyExists",
    "CommandErrorEvent",
    "CommandInvocationError",
    "CommandIsOnCooldown",
    "CommandLike",
    "CommandNotFound",
    "Context",
    "ConverterFailure",
    "CooldownAlgorithm",
    "CooldownManager",
    "CooldownStatus",
    "DefaultHelpCommand",
    "EmojiConverter",
    "ExtensionAlreadyLoaded",
    "ExtensionMissingLoad",
    "ExtensionMissingUnload",
    "ExtensionNotFound",
    "ExtensionNotLoaded",
    "FixedWindowCooldownAlgorithm",
    "GlobalBucket",
    "GuildBucket",
    "GuildCategoryConverter",
    "GuildChannelConverter",
    "GuildConverter",
    "GuildVoiceChannelConverter",
    "HumanOnly",
    "InsufficientCache",
    "InviteConverter",
    "LightbulbError",
    "LightbulbEvent",
    "LightbulbStartedEvent",
    "MemberConverter",
    "MessageCommand",
    "MessageCommandCompletionEvent",
    "MessageCommandErrorEvent",
    "MessageCommandInvocationEvent",
    "MessageContext",
    "MessageConverter",
    "MissingRequiredPermission",
    "MissingRequiredRole",
    "NSFWChannelOnly",
    "NotEnoughArguments",
    "NotOwner",
    "OnlyInDM",
    "OnlyInGuild",
    "OptionLike",
    "OptionModifier",
    "OptionsProxy",
    "Parser",
    "Plugin",
    "PrefixCommand",
    "PrefixCommandCompletionEvent",
    "PrefixCommandErrorEvent",
    "PrefixCommandGroup",
    "PrefixCommandInvocationEvent",
    "PrefixContext",
    "PrefixGroupMixin",
    "PrefixSubCommand",
    "PrefixSubGroup",
    "ResponseProxy",
    "RoleConverter",
    "SlashCommand",
    "SlashCommandCompletionEvent",
    "SlashCommandErrorEvent",
    "SlashCommandGroup",
    "SlashCommandInvocationEvent",
    "SlashContext",
    "SlashGroupMixin",
    "SlashSubCommand",
    "SlashSubGroup",
    "SlidingWindowCooldownAlgorithm",
    "SnowflakeConverter",
    "SubCommandTrait",
    "TextableGuildChannelConverter",
    "TimestampConverter",
    "UserBucket",
    "UserCommand",
    "UserCommandCompletionEvent",
    "UserCommandErrorEvent",
    "UserCommandInvocationEvent",
    "UserContext",
    "UserConverter",
    "WebhookOnly",
    "add_checks",
    "add_cooldown",
    "app",
    "app_command_permissions",
    "bot_has_channel_permissions",
    "bot_has_guild_permissions",
    "bot_has_role_permissions",
    "bot_only",
    "check_exempt",
    "buckets",
    "checks",
    "command",
    "commands",
    "context",
    "converters",
    "cooldowns",
    "cooldown_algorithms",
    "decorators",
    "dm_only",
    "errors",
    "events",
    "filter_commands",
    "guild_only",
    "has_attachments",
    "has_channel_permissions",
    "has_guild_permissions",
    "has_role_permissions",
    "has_roles",
    "help_command",
    "human_only",
    "implements",
    "nsfw_channel_only",
    "option",
    "owner_only",
    "parser",
    "plugins",
    "set_help",
    "set_max_concurrency",
    "utils",
    "webhook_only",
    "when_mentioned_or",
]

from lightbulb import app
from lightbulb import buckets
from lightbulb import checks
from lightbulb import commands
from lightbulb import context
from lightbulb import converters
from lightbulb import cooldown_algorithms
from lightbulb import cooldowns
from lightbulb import decorators
from lightbulb import errors
from lightbulb import events
from lightbulb import help_command
from lightbulb import parser
from lightbulb import plugins
from lightbulb import utils
from lightbulb.app import *
from lightbulb.buckets import *
from lightbulb.checks import *
from lightbulb.commands import *
from lightbulb.context import *
from lightbulb.converters import *
from lightbulb.cooldown_algorithms import *
from lightbulb.cooldowns import *
from lightbulb.decorators import *
from lightbulb.errors import *
from lightbulb.events import *
from lightbulb.help_command import *
from lightbulb.parser import *
from lightbulb.plugins import *

__version__ = "2.3.3"



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["get", "find"]

import typing as t
from operator import attrgetter

T = t.TypeVar("T")


def get(sequence: t.Iterable[T], **attrs: t.Any) -> t.Optional[T]:
    """
    Get the first item from an iterable that matches all the parameters
    specified, or return ``None`` if no matching item was found.

    Args:
        sequence (Iterable[ T ]): Iterable to search through.

    Keyword Args:
        **attrs: Attributes to match.

    Example:
        Searching for a member with a specific username in the bot's cached members.

        .. code-block:: python

            members = bot.get_members_view_for_guild(ctx.guild_id)
            member = lightbulb.utils.get(members.values(), username="foo")

    See Also:
        :obj:`~find`
    """
    flattened = [(attrgetter(attr), value) for attr, value in attrs.items()]

    for item in sequence:
        if all([getter(item) == value for getter, value in flattened]):
            return item
    return None


def find(sequence: t.Iterable[T], predicate: t.Callable[[T], bool]) -> t.Optional[T]:
    """
    Find the first item from an iterable that passes for the predicate specified,
    or return ``None`` if no matching item was found.

    Args:
        sequence (Iterable[ T ]): Iterable to search through.
        predicate (Callable[ [ T ], :obj:`bool` ]): Function to evaluate if the item
            is the correct one or not. It should return a boolean or boolean-like result.

    Example:
        Searching for a member with a specific username in the bot's cached members.

        .. code-block:: python

            members = bot.get_members_view_for_guild(ctx.guild_id)
            member = lightbulb.utils.find(members.values(), lambda m: m.username == "foo")

    See Also:
        :obj:`~get`
    """
    for item in sequence:
        if predicate(item):
            return item
    return None



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
import os
import re
import types

from setuptools import dist
from setuptools import find_namespace_packages
from setuptools import setup
from setuptools.command import install

name = "lightbulb"


def parse_meta():
    with open(os.path.join(name, "__init__.py")) as fp:
        code = fp.read()

    token_pattern = re.compile(r"^__(?P<key>\w+)?__\s*=\s*(?P<quote>(?:'{3}|\"{3}|'|\"))(?P<value>.*?)(?P=quote)", re.M)

    groups = {}

    for match in token_pattern.finditer(code):
        group = match.groupdict()
        groups[group["key"]] = group["value"]

    return types.SimpleNamespace(**groups)


def long_description():
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path):
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]


meta = parse_meta()

setup(
    name="hikari-lightbulb",
    version=meta.version,
    description="A simple to use command handler for Hikari",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author="tandemdude",
    author_email="tandemdude1@gmail.com",
    url="https://github.com/tandemdude/hikari-lightbulb",
    packages=find_namespace_packages(include=[name + "*"]),
    license="LGPL-3.0-ONLY",
    include_package_data=True,
    zip_safe=False,
    install_requires=parse_requirements_file("requirements.txt"),
    extras_require={"crontrigger": parse_requirements_file("crontrigger_requirements.txt")},
    python_requires=">=3.8.0,<3.12",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["BaseConverter"]

import abc
import typing as t

if t.TYPE_CHECKING:
    from lightbulb import context as context_

T = t.TypeVar("T")


class BaseConverter(abc.ABC, t.Generic[T]):
    """
    Base converter class

    Args:
        context (:obj:`~.context.base.Context`): Context to convert the argument under.
    """

    __slots__ = ("context",)

    def __init__(self, context: context_.base.Context) -> None:
        self.context: context_.Context = context
        """Context to convert the argument(s) under."""

    @abc.abstractmethod
    async def convert(self, arg: str) -> T:
        """
        Conversion method for the converter.

        Args:
            arg (:obj:`str`): Argument to convert.

        Returns:
            T: Converted argument.
        """
        ...



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "BaseConverter",
    "BooleanConverter",
    "ColorConverter",
    "ColourConverter",
    "EmojiConverter",
    "GuildCategoryConverter",
    "GuildChannelConverter",
    "GuildConverter",
    "GuildVoiceChannelConverter",
    "InviteConverter",
    "MemberConverter",
    "MessageConverter",
    "RoleConverter",
    "SnowflakeConverter",
    "TextableGuildChannelConverter",
    "TimestampConverter",
    "UserConverter",
]

import datetime
import typing as t

import hikari

from lightbulb.converters import base
from lightbulb.converters import special
from lightbulb.converters.base import *
from lightbulb.converters.special import *

CONVERTER_TYPE_MAPPING: t.Mapping[t.Any, t.Type[base.BaseConverter[t.Any]]] = {
    hikari.User: special.UserConverter,
    hikari.Member: special.MemberConverter,
    hikari.GuildChannel: special.GuildChannelConverter,
    hikari.TextableGuildChannel: special.TextableGuildChannelConverter,
    hikari.TextableChannel: special.TextableGuildChannelConverter,
    hikari.GuildCategory: special.GuildCategoryConverter,
    hikari.GuildVoiceChannel: special.GuildVoiceChannelConverter,
    hikari.Role: special.RoleConverter,
    hikari.Emoji: special.EmojiConverter,
    hikari.Guild: special.GuildConverter,
    hikari.Message: special.MessageConverter,
    hikari.Invite: special.InviteConverter,
    hikari.Colour: special.ColourConverter,
    hikari.Color: special.ColourConverter,
    hikari.Snowflake: special.SnowflakeConverter,
    datetime.datetime: special.TimestampConverter,
}
